import subprocess
import time
import os
import json
import uuid
from datetime import datetime
import sys
import shutil

from curie.logger import init_logger, send_question_telemetry

# Constants
DEFAULT_TASK_CONFIG = {
    "job_name": "default_research",
    "docker_image": "curie-pip-image",
    "dockerfile_name": "ExpDockerfile_pip", 
    "benchmark_specific_context": "none",
    "is_user_interrupt_allowed": False,
    "timeout": 600,
    "max_coding_iterations": 25,
    "supervisor_system_prompt_filename": "prompts/simple/simple-supervisor.txt",
    "control_worker_system_prompt_filename": "prompts/simple/simple-control-worker.txt",
    "patcher_system_prompt_filename": "prompts/simple/simple-patcher.txt",
    "llm_verifier_system_prompt_filename": "prompts/simple/simple-llm-verifier.txt",
    "coding_prompt_filename": "prompts/simple/simple-coding.txt",
    "worker_system_prompt_filename": "prompts/simple/simple-worker.txt",
    "workspace_name": "", # meant to be empty
    "dataset_dir": "", # meant to be empty
}

DEFAULT_JOB_NAME = "default_research"

def write_api_keys_to_env(api_keys: dict):
    """Write API keys to env.sh file."""
    env_path = os.path.join(os.getcwd(), '.setup', 'env.sh')
    os.makedirs(os.path.dirname(env_path), exist_ok=True)
    # print(f"Writing API keys to {env_path}")
    
    with open(env_path, 'w') as f:
        for key, value in api_keys.items():
            print(f"Writing {key} to {env_path}")
            f.write(f'export {key}="{value}"\n')

def docker_image_exists(image):
    """Check if a Docker image exists locally."""
    try:
        result = subprocess.run(
            ["docker", "image", "inspect", image], 
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )
        return result.returncode == 0
    except Exception as e:
        print(f"Error checking Docker image: {e}")
        return False

def build_docker_image(image_name, dockerfile):
    """Build Docker image if it doesn't exist."""
    command = [
        "sudo", "docker", "build",
        "--no-cache", "--progress=plain",
        "-t", image_name,
        "-f", dockerfile,
        "."
    ]
    subprocess.run(command, check=True)

def run_docker_container(unique_id, iteration, task_config, logger):
    """Run a Docker container for the experiment."""
    rand_uuid = uuid.uuid4()
    container_name = f"exp-agent-container-{unique_id}-{rand_uuid}-iter_{iteration}"
    
    image_name = task_config["docker_image"]
    docker_filename = task_config["base_dir"] + "/curie/" + task_config["dockerfile_name"]

    if docker_image_exists(image_name):
        logger.info(f"Using existing Docker image: {image_name}")
    else:
        logger.info(f"Start building Docker image {image_name} ... ")
        build_docker_image(image_name, docker_filename)
    
    base_dir = os.getcwd()
    api_key_dir = os.path.join(os.getcwd(), '.setup')
    command = [
        "docker", "run",
        "-v", "/var/run/docker.sock:/var/run/docker.sock",
        "-v", f"{api_key_dir}:/curie/setup/:ro",
        "-v", f"{base_dir}/logs:/logs",
        "-v", f"{base_dir}/workspace:/workspace",
        "-v", f"/:/all:ro",
        "--network=host",
        "-d",
    ]
    
    # Add GPU support if available
    has_gpu = shutil.which("nvidia-smi") is not None and subprocess.call(
        ["nvidia-smi"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0
    if has_gpu:
        command += ["--gpus", "all"]
        
    command += ["--name", container_name, image_name]

    logger.info(f"Running command: {' '.join(command)}")
    subprocess.run(command, check=True) 
    return container_name

def execute_experiment_in_container(container_name, config_file, logger):
    """Execute the experiment inside the Docker container."""
    logger.info(f"Starting experiment in container {container_name} with config in {config_file}")
            
    organization_id = os.environ.get("ORGANIZATION") if os.environ.get("ORGANIZATION") else "014482"
    # Command to run inside container
    container_command = (
        "source setup/env.sh && "
        '''eval "$(micromamba shell hook --shell bash)" && '''
        "micromamba activate curie && "
        f"sed -i '474i \\                    \"organization\": \"{organization_id}\",' /root/.cache/pypoetry/virtualenvs/openhands-ai-*-py3.12/lib/python3.12/site-packages/litellm/llms/azure/azure.py &&"
        f"sed -i '474i \\    \"organization\": \"{organization_id}\",' /opt/micromamba/envs/curie/lib/python3.11/site-packages/litellm/llms/azure/azure.py  &&"
        "sed -i '49d' /root/.cache/pypoetry/virtualenvs/openhands-ai-*-py3.12/lib/python3.12/site-packages/litellm/llms/azure/chat/o_series_handler.py &&"
        f"sed -i '49i \\                    organization=\"{organization_id}\",' /root/.cache/pypoetry/virtualenvs/openhands-ai-*-py3.12/lib/python3.12/site-packages/litellm/llms/azure/chat/o_series_handler.py  &&"
        f"python3 construct_workflow_graph.py /{config_file}"
    )
    
    try:
        subprocess.run([
            "docker", "exec", "-it", container_name,
            "bash", "-c", container_command
        ], check=True)
        return True
    except subprocess.CalledProcessError as e:
        logger.error(f"Experiment failed with exit code {e.returncode}. Error: {e}")
        return False

def cleanup_docker_container(container_name):
    """Stop and remove the Docker container."""
    try:
        print(f"Stopping and removing Docker container: {container_name}...")
        subprocess.run(["docker", "stop", container_name], check=True)
        subprocess.run(["docker", "rm", container_name], check=True)
        print(f"Docker container {container_name} cleaned up.")
    except subprocess.SubprocessError as e:
        print(f"Error cleaning up container: {e}")

def run_prune_commands():
    """Run Docker pruning commands to free up resources."""
    commands = [
        ["docker", "container", "prune", "-f"],
        ["docker", "image", "prune", "-f"],
        ["docker", "volume", "prune", "-f"],
        ["docker", "builder", "prune", "-f"],
    ]

    for command in commands:
        try:
            print(f"Running docker: {' '.join(command)}")
            result = subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            print(result.stdout.decode())
        except subprocess.CalledProcessError as e:
            print(f"Error running command: {' '.join(command)}")
            print(e.stderr.decode())

def get_workspace_name(task_config):
    """Extract workspace name from task config."""
    return (
        (os.path.basename(task_config.get('workspace_name', '')) or '' ) or 
        task_config.get('job_name', '') or 
        DEFAULT_JOB_NAME
    )

def create_config_file(question_file, unique_id, iteration, task_config):
    """Create experiment configuration file and set up logging."""
    work_name = get_workspace_name(task_config)
    
    # Setup logging directory and files
    exp_log_dir = os.path.join("logs", f"{work_name}_{unique_id}_iter{iteration}")
    os.makedirs(exp_log_dir, exist_ok=True)

    # Generate filenames
    question_base = os.path.basename(question_file).replace('.txt', '')
    log_filename = os.path.join(exp_log_dir, f"{question_base}_{unique_id}_iter{iteration}.log")
    config_filename = os.path.join(exp_log_dir, 
                                f"{work_name}_config_{question_base}_{unique_id}_iter{iteration}.json")

    # Update task configuration
    base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    task_config.update({
        "unique_id": unique_id,
        "iteration": iteration,
        "log_filename": log_filename,
        "exp_plan_filename": question_file,
        "base_dir": base_dir,
    })
        
    os.makedirs(os.path.dirname(config_filename), exist_ok=True)
    send_question_telemetry(question_file)
    
    with open(config_filename, "w") as f:
        json.dump(task_config, f, indent=4)
    send_question_telemetry(config_filename)
    
    logger = init_logger(log_filename)
    logger.info(f"Config file created: {config_filename}")
    logger.info(f"Check out the log file: {log_filename}")
    
    return task_config, config_filename, logger

def prepare_question_file(task_config, question_text):
    """Create a question file from question text."""
    q_file = get_workspace_name(task_config)
    question_file = f'workspace/{q_file}_{int(time.time())}.txt'
    
    try:
        os.makedirs(os.path.dirname(question_file), exist_ok=True)
        with open(question_file, 'w') as f:
            f.write(question_text)
        return question_file
    except Exception as e:
        print(f"Error writing question to file: {e}")
        print("Please give permission to write to `workspace/`.")
        sys.exit(1)

def execute_curie(question_filename, unique_id, iteration, task_config):
    """Execute a single Curie iteration."""
    # Create configuration file and get logger
    task_config, config_filename, logger = create_config_file(
        question_filename, unique_id, iteration, task_config)

    # Run Docker container for this iteration
    container_name = None
    try:
        container_name = run_docker_container(unique_id, iteration, task_config, logger)
        execute_experiment_in_container(container_name, config_filename, logger)
    finally:
        # Clean up Docker container after each iteration
        if container_name:
            cleanup_docker_container(container_name)
        run_prune_commands()
    
    send_question_telemetry(task_config['log_filename'])

def update_config(task_config, workspace_name=None, dataset_dir=None, max_global_steps=30):
    """Load and update task configuration with command line arguments."""
    # check if workspace_name is a valid path
    if workspace_name and not os.path.exists(workspace_name):
        raise ValueError(f"Workspace name {workspace_name} is not a valid path.")
    # check if dataset_dir is a valid path
    if dataset_dir and not os.path.exists(dataset_dir):
        raise ValueError(f"Dataset directory {dataset_dir} is not a valid path.") 
    
    if task_config is None:
        task_config = DEFAULT_TASK_CONFIG
        task_config['workspace_name'] = workspace_name or task_config['workspace_name']
        task_config['dataset_dir'] = dataset_dir or task_config['dataset_dir']
        return task_config

    task_config['workspace_name'] = workspace_name or task_config['workspace_name']
    task_config['dataset_dir'] = dataset_dir or task_config['dataset_dir']
    # fill up the unspecified fields in the task config with DEFAULT_TASK_CONFIG
    for key, value in DEFAULT_TASK_CONFIG.items():
        if key not in task_config:
            task_config[key] = value
    # force override the docker image and dockerfile name
    task_config['max_global_steps'] = max_global_steps
    task_config['docker_image'] = "curie-pip-image"
    task_config['dockerfile_name'] = "ExpDockerfile_pip" 
    return task_config

def validate_question_input(question_file, question):
    """Validate that exactly one of question_file or question is provided."""
    if question_file is None and question is None:
        print("Please provide either a question file or a question.")
        return False
    elif question_file is not None and question is not None:
        print("Please provide only one of either a question file or a question.")
        return False
    return True

def experiment(api_keys=None, dataset_dir=None, workspace_name=None, question_file=None, question=None, iterations=1, task_config=None, max_global_steps=30):
    """Main experiment function that orchestrates the experiment workflow."""
    # Write API keys to env file if provided
    if api_keys:
        write_api_keys_to_env(api_keys)
    
    # Load and update configuration
    task_config = update_config(task_config, workspace_name, dataset_dir, max_global_steps)
    
    print(f"Curie is running with the following configuration: {task_config}")
    
    # Validate question input
    if not validate_question_input(question_file, question):
        return
    
    # Prepare question file
    question_file = question_file if question_file else prepare_question_file(task_config, question)
    
    # Run iterations
    for iteration in range(1, iterations + 1):
        start_time = time.time()
        unique_id = datetime.now().strftime("%Y%m%d%H%M%S")
        
        execute_curie(question_file, unique_id, iteration, task_config)
        
        end_time = time.time()
        elapsed_time = end_time - start_time
        print(f"Iteration {iteration} for {question_file} completed in {elapsed_time:.2f} seconds.")




NUM_WORKERS = 1
NUM_CONTROL_WORKERS = 1

# Scheduler specific:
VARS_PER_PARTITION = 5 # Number of variable values to assign to each partition. This means we may partition a single experimental plan into multiple smaller plans that are identical except for the experimental/control groups used. E.g., [{instance_type: t2.micro}, {instance_type: t2.small}, {instance_type: t2.medium}, {instance_type: t2.large}, {instance_type: t2.xlarge}] would be 5 variable values. 
PARTITIONS_PER_WORKER = 1 # Number of experimental/control group partitions to assign to one worker. Note this is only used later after the controlled experiment has been set up correctly. 

def list_worker_names() -> list:
    worker_names = []
    for i in range(NUM_WORKERS):
        worker_names.append(f"worker_{i}")
    return worker_names

def list_control_worker_names() -> list:
    worker_names = []
    for i in range(NUM_CONTROL_WORKERS):
        worker_names.append(f"control_worker_{i}")
    return worker_names

AGENT_LIST = ["supervisor"]
AGENT_LIST.extend(list_worker_names()) # add worker agents to the agent list
AGENT_LIST.extend(list_control_worker_names()) # add control worker agents to the agent list

# Do not override the following variables:
CONCLUDER_BUFFER_STEPS = 15


import subprocess
import time
import os
import shutil  # Import for deleting directories
import psutil


def clear_workspace():
    """
    Deletes all contents of the /workspace directory but keeps the directory itself.
    """
    workspace_path = "/workspace"
    if os.path.exists(workspace_path):
        try:
            # Iterate over all contents in the directory
            for item in os.listdir(workspace_path):
                item_path = os.path.join(workspace_path, item)
                if os.path.isfile(item_path) or os.path.islink(item_path):
                    os.unlink(item_path)  # Remove files or symlinks
                elif os.path.isdir(item_path):
                    shutil.rmtree(item_path)  # Remove directories
            print(f"Cleared all contents of the directory: {workspace_path}")
        except Exception as e:
            print(f"Error clearing contents of {workspace_path}: {e}")
    else:
        print(f"Directory {workspace_path} does not exist. Nothing to clear.")

def run_command_x_times(base_command, log_prefix, config_file, x, timeout):
    """
    Runs a shell command X number of times with dynamic log file names and a fixed configuration file.
    Skips to the next iteration if the log file does not receive new lines within the specified timeout.

    :param base_command: The base command to execute (excluding the log file and config file).
    :param log_prefix: The prefix for the log file names.
    :param config_file: The configuration file to pass as the second argument.
    :param x: The number of times to run the command.
    :param timeout: The duration (in seconds) to wait for new lines in the log file.
    """
    for i in range(1, x + 1):  # Iteration numbers start from 1
        # log_file = f"{log_prefix}{i}.log"  # Generate log file name
        command = f"{base_command}"  # Complete command
        print(f"Running command iteration {i}: {command}")
        
        # Start the command
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        
        # Monitor the log file for changes
        start_time = time.time()
        last_size = 0
        
        while True:
            if time.time() - start_time > timeout:
                print(f"No new lines detected in for {timeout} seconds. Moving to the next iteration.")
                process.kill()
                process.wait()  # Ensure the process is completely terminated
                print("Process forcefully terminated.")
                break  # Exit the loop and move to the next iteration
            
            time.sleep(1)  # Check the file every second
            
            # Check if the process has completed
            if process.poll() is not None:
                print(f"Command finished for iteration {i}.")
                break

        # Verify that process has closed:
        # Get the process ID (PID)
        pid = process.pid

        # Check if the process is still running
        if psutil.pid_exists(pid):
            print(f"Process with PID {pid} is still running...")
        else:
            print(f"Process with PID {pid} has terminated.")
        
        # # Clear the /workspace folder. This is used by the control and experimental workers to save their files. 
        # clear_workspace()

if __name__ == "__main__":
    # Base command without log file or config file
    base_command = "sleep 10"
    
    # Log file prefix
    log_prefix = "logs/log-temp"
    
    # Configuration file
    config_file = "configs/cloud_config.json"  # Replace with your desired config file
    
    # Number of times to execute the command
    iterations = 10  # Replace with your desired number
    
    # Timeout in seconds for monitoring log file
    timeout = 9  # Replace with your desired timeout duration
    
    # Run the command
    run_command_x_times(base_command, log_prefix, config_file, iterations, timeout)



import tool

command = ''' 
cat << EOT > junktest.sh
CPU_UTILIZATION=$(ssh -o StrictHostKeyChecking=no -i $KEY_PATH ec2-user@$PUBLIC_IP 'top -bn1 | grep "Cpu(s)" | sed "s/.*, *\([0-9.]*\)%* id.*/\1/" | awk '{print 100 - $1"%"}')
EOT
'''

print(tool.execute_shell_command(command))


# TEST 1: test through regular langchain tool invoke:
import pytest
from pydantic import ValidationError
from typing import Annotated, Literal
from typing_extensions import TypedDict
from langgraph.graph.message import add_messages

import scheduler as sched
import tool
from langgraph.store.memory import InMemoryStore

class State(TypedDict):
    # Messages have the type "list". The `add_messages` function
    # in the annotation defines how this state key should be updated
    # (in this case, it appends messages to the list, rather than overwriting them)
    messages: Annotated[list, add_messages]

state_instance: State = {
    "messages": []
}

store = InMemoryStore()
metadata_store = InMemoryStore()

store_write_tool = tool.StoreWriteTool(store, metadata_store)
sched.setup_sched(metadata_store)

def test_invalid_plan_input(store_write_tool):
    """
    Test that invoking store_write_tool with an invalid 'plan' structure raises an error.
    """
    invalid_input = {
        "plan": {"testkey": "University of Michigan!"}  # Missing required keys from the schema
    }

    # Expect a ValidationError when invoking with invalid input
    with pytest.raises(ValidationError) as exc_info:
        memory_id = store_write_tool.invoke(invalid_input)

    print("Test passed: Invalid input raised a ValidationError.")

def test_valid_plan_input(store_write_tool):
    """
    Test that invoking store_write_tool with a valid 'plan' structure returns a memory_id.
    """
    valid_input = {
        "plan": {
            "hypothesis": "AWS EC2 VMs in us-east-1 are slower than us-east-2.",
            "constant_vars": ["EC2 VM type"],
            "independent_vars": ["AWS region"],
            "dependent_vars": ["execution_time"],
            "controlled_experiment_setup_description": "Create EC2 VM, SSH into VM, run predefined task.",
            "control_group": {"region": ["us-east-1"]},
            "experimental_group": {"region": ["us-east-2"]},
            "priority": 1,
        }
    }

    memory_id = store_write_tool.invoke(valid_input, state_instance)
    store_get_tool = tool.StoreGetTool(store)
    vals = store_get_tool.invoke({"plan_key": memory_id})
    assert memory_id is not None and vals is not None

    print("Test passed: Valid input returned a memory_id.")

def test_multi_valid_plan_input(store_write_tool):
    """
    Test that invoking store_write_tool with a valid 'plan' structure, followed by an edit of the written plan (note: with non-validated input keys), works.
    """
    valid_input = {
        "plan": {
            "hypothesis": "AWS EC2 VMs in us-east-1 are slower than us-east-2.",
            "constant_vars": ["EC2 VM type"],
            "independent_vars": ["AWS region"],
            "dependent_vars": ["execution_time"],
            "controlled_experiment_setup_description": "Create EC2 VM, SSH into VM, run predefined task.",
            "control_group": {"region": ["us-east-1"]},
            "experimental_group": {"region": ["us-east-2"]},
            "priority": 1,
        }
    }

    memory_id = store_write_tool.invoke(valid_input)
    store_get_tool = tool.StoreGetTool(store)
    vals = store_get_tool.invoke({"plan_id": memory_id})

    valid_input = {
        "plan": {
            "hypothesis": "AWS EC2 VMs in us-east-1 are slower than us-east-2.",
            "constant_vars": ["EC2 VM type"],
            "independent_vars": ["AWS region"],
            "dependent_vars": ["execution_time"],
            "controlled_experiment_setup_description": "Create EC2 VM, SSH into VM, run predefined task.",
            "control_group": {"region": ["us-east-1"]},
            "experimental_group": {"region": ["us-east-2"]},
            "priority": 1,
            "control_group_done": True, # new key value
        },
        "plan_id": memory_id
    }
    # print("entering edit")
    new_memory_id = store_write_tool.invoke(valid_input)
    assert new_memory_id == memory_id
    store_get_tool = tool.StoreGetTool(store)
    vals = store_get_tool.invoke({"plan_id": new_memory_id})
    print("New plan:")
    print(vals)

    assert memory_id is not None and vals is not None

    assert len(vals) == 1
    assert vals[0]["control_group_done"] == True

    print("Test passed: Valid input returned a memory_id, and editing the memory_id with a new plan worked.")

test_invalid_plan_input(store_write_tool)
test_valid_plan_input(store_write_tool)
test_multi_valid_plan_input(store_write_tool)

# TEST 2: Store through agent test:
import store_agent
from typing import Annotated

from typing_extensions import TypedDict

from langgraph.graph import StateGraph, START, END
from langgraph.graph.message import add_messages
from langgraph.prebuilt import ToolNode, tools_condition
from langchain_core.messages import HumanMessage, SystemMessage

from langgraph.store.memory import InMemoryStore

import model
import utils
import tool

class State(TypedDict):
    # Messages have the type "list". The `add_messages` function
    # in the annotation defines how this state key should be updated
    # (in this case, it appends messages to the list, rather than overwriting them)
    messages: Annotated[list, add_messages]

graph_builder = StateGraph(State)

store = InMemoryStore()
metadata_store = InMemoryStore()
store_write_tool = tool.StoreWriteTool(store, metadata_store)
sched.setup_sched(metadata_store)
store_get_tool = tool.StoreGetTool(store)
tools = [store_write_tool, 
    store_get_tool
]
# system_prompt_file = "prompts/exp-store-writer.txt"
system_prompt_file = "prompts/exp-store-writer-test.txt"
store_agent = store_agent.create_StoreWriteAgent(tools, system_prompt_file, State)

graph_builder.add_node("chatbot", store_agent)
tool_node = ToolNode(tools=tools)
graph_builder.add_node("tools", tool_node)

graph_builder.add_edge(START, "chatbot")
graph_builder.add_conditional_edges(
    "chatbot",
    tools_condition,
)
graph_builder.add_edge("tools", "chatbot")
# graph_builder.add_edge("chatbot", END)
graph = graph_builder.compile()

utils.save_langgraph_graph(graph, "misc/graph_image.png") 

# Section: Run agent
def stream_graph_updates(user_input: str):
    for event in graph.stream({"messages": [("user", user_input)]}):
        for value in event.values():
            print("Assistant:", value["messages"][-1].content)

while True:
    try:
        user_input = input("User: ")
        if user_input.lower() in ["quit", "exit", "q"]:
            print("Goodbye!")
            break

        stream_graph_updates(user_input)
    except:
        # fallback if input() is not available
        user_input = "What do you know about LangGraph?"
        print("User: " + user_input)
        stream_graph_updates(user_input)
        break


import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import io
import json
import re
import ast
import os
import re
import toml
import shutil
import subprocess
from pathlib import Path
from typing import List, Dict, Any

def get_all_price_per_1k_tokens() -> Dict[str, Dict[str, float]]:
    return {
        "gpt-4o": {"input": 0.0025, "output": 0.01}, 
        "gpt-4o-mini": {"input": 0.00015, "output": 0.000075},
        "anthropic.claude-3-7-sonnet-20250219-v1:0": {"input": 0.003, "output": 0.015},
    }

def get_model_context_length() -> int:
    """Get the context length for the current model."""
    # FIXME: add more models as needed
    context_length_dict = {
        "gpt-4o": 128000,
        "gpt-4o-mini": 128000,
        "anthropic.claude-3-7-sonnet-20250219-v1:0": 200000,
    }
    model_name = get_model_name()
    return context_length_dict.get(model_name, 30000)

def get_input_price_per_token() -> float:
    """Get the price per token for input text."""
    model_name = get_model_name()
    return get_all_price_per_1k_tokens()[model_name]["input"] / 1000

def get_output_price_per_token() -> float:
    """Get the price per token for output text."""
    model_name = get_model_name()
    return get_all_price_per_1k_tokens()[model_name]["output"] / 1000

def get_model_name() -> str:
    """Strip provider prefix if present (e.g., "openai/gpt-4" -> "gpt-4")"""
    current_model = os.environ.get("MODEL", "gpt-4o")
    # Strip provider prefix if present (e.g., "openai/gpt-4" -> "gpt-4")
    model_name = current_model.split('/')[-1]
    if "claude" in model_name and "us." in model_name: # example: "us.anthropic.claude-3-7-sonnet-20250219-v1:0"
        # Remove "us." prefix:
        model_name = model_name.split("us.")[1]
    return model_name

def extract_plan_id(prompt: str) -> str:
    """
    Extracts the plan ID from the given prompt.

    Args:
        prompt (str): The input text containing a plan ID.

    Returns:
        str: The extracted plan ID if found, else an empty string.
    """
    # Regular expression to match UUID-like patterns (plan_id format)
    pattern = r"\b[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}\b"

    # Search for the pattern in the prompt
    match = re.search(pattern, prompt)

    if match:
        return True
    else:
        return False
    
    # # Return the matched plan ID if found, else return an empty string
    # return match.group(0) if match else ""

def extract_partition_name(prompt: str) -> str:
    """
    Extracts the partition name from a given prompt.

    Args:
        prompt (str): The input text that may contain a partition name.

    Returns:
        str: The extracted partition name if found, else an empty string.
    """
    # Regular expression to match partition names (e.g., 'partition_1')
    pattern = r"['\"]?(partition_\d+)['\"]?"
    
    # Search for the pattern in the prompt
    match = re.search(pattern, prompt)

    if match:
        return True
    else:
        return False
    
    # Return the matched partition name if found, else return an empty string
    # return match.group(1) if match else ""

def extract_workspace_dir(text: str) -> str:
    """
    Extracts the directory name that appears after '/workspace/' in the given text,
    excluding any surrounding single quotes.

    Args:
        text (str): The input string containing the workspace path.

    Returns:
        str: The directory name after '/workspace/', or an empty string if not found.
    """
    match = re.search(r"/workspace/([^\s'/]+)", text)  # Excludes single quotes
    if match:
        return True
    else:
        return False
    # return match.group(1) if match else ""

def check_file_exists(file_path: str) -> bool:
    """
    Checks if a file exists at the given path.

    Args:
        file_path (str): The path to the file.

    Returns:
        bool: True if the file exists, False otherwise.
    """
    return os.path.isfile(file_path)

def print_workspace_contents():
    workspace_dir = "/workspace"

    if os.path.exists(workspace_dir):
        print(f"Contents of {workspace_dir}:")
        for root, dirs, files in os.walk(workspace_dir):
            print(f"Root: {root}")
            if dirs:
                print(f"Directories: {dirs}")
            if files:
                print(f"Files: {files}")
            print("-" * 40)  # Separator for clarity
    else:
        print(f"Directory {workspace_dir} does not exist.")

def save_langgraph_graph(graph, dst_filename) -> None:
    try:
        # Generate the graph as a PNG binary
        graph_png = graph.get_graph().draw_mermaid_png()
        # Convert binary data to an image using Matplotlib
        img = mpimg.imread(io.BytesIO(graph_png), format='png')
        plt.imshow(img)
        plt.axis('off')  # Hide axes for a cleaner display
        plt.savefig(dst_filename, dpi=300, bbox_inches='tight')  # Save the image with high quality
        plt.close()  # Close the plot to avoid overlapping when running multiple saves
    except Exception as e:
        print(f"Error displaying graph with Matplotlib: {e}")

def pretty_json(obj):
    return json.dumps(obj, sort_keys=True, indent=4, default=str)

def parse_nested(value):
    """
    Parse a value that could be a nested structure (dict, list, etc.).
    Uses ast.literal_eval for safe evaluation of Python-like literals.
    """
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError):
        return value  # Return raw value if parsing fails


def extract_key_value_pairs(input_string):
    """
    Extracts all top-level key-value pairs from the input string.
    Handles nested dictionaries and lists recursively.
    """
    # Split input by top-level keys (matches <key>=<value>)
    pattern = r"(\w+)=((?:'[^']*'|\{.*?\}|\[.*?\]))"
    matches = re.finditer(pattern, input_string, re.DOTALL)

    result = {}
    last_end = 0

    for match in matches:
        key, value = match.groups()
        key = key.strip()
        value = value.strip()

        # Check for truncated values (e.g., nested dictionaries) and expand them
        if value.startswith("'") and value.endswith("'"):
            # Simple string value
            parsed_value = value[1:-1]
        elif value.startswith("{") or value.startswith("["):
            # Potential nested structure
            # Capture everything between matched braces or brackets
            balance_count = 0
            start = match.start(2)  # Start of value in the string
            for i, char in enumerate(input_string[start:], start=start):
                if char == '{' or char == '[':
                    balance_count += 1
                elif char == '}' or char == ']':
                    balance_count -= 1
                if balance_count == 0:
                    # Found the complete nested structure
                    full_value = input_string[start:i + 1]
                    parsed_value = parse_nested(full_value)
                    last_end = i + 1
                    break
        else:
            # Fallback for unstructured or invalid values
            parsed_value = value

        result[key] = parsed_value

    return result

def parse_langchain_llm_output(input_string):
    try: # https://python.langchain.com/api_reference/core/messages/langchain_core.messages.ai.AIMessage.html#langchain_core.messages.ai.AIMessage.pretty_print
        return input_string.pretty_print()
    except Exception as e:
        try: 
            structured_data = extract_key_value_pairs(input_string)
            return json.dumps(structured_data, indent=4)
        except Exception as e:
            return f"Error parsing LangChain LLM output. \nError: {e}. \nRaw output: {input_string}"

def parse_env_string(env_string):
    """Parse environment string and return a dictionary of key-value pairs."""
    env_vars = {}
    for line in env_string.splitlines():
        # Skip empty lines or comment-only lines
        line = line.strip()
        if not line or line.startswith('#'):
            continue
            
        # Remove 'export' if present
        if line.startswith('export'):
            line = line.replace('export', '', 1).strip()
            
        # Split on first '=' and handle inline comments
        if '=' in line:
            # Split on first '#' to remove comments
            line_without_comment = line.split('#')[0].strip()
            
            # Now split on first '=' to get key-value
            key, value = line_without_comment.split('=', 1)
            
            # Clean up key and value
            key = key.strip()
            value = value.strip().strip('"\'')
            
            if key:  # Only add if key is not empty
                env_vars[key] = value
                
    return env_vars

def categorize_variables(env_vars):
    """Categorize environment variables into config sections."""
    has_gpu = shutil.which("nvidia-smi") is not None and subprocess.call(["nvidia-smi"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0

    config = {
        'core': {
            'file_store': 'local',
            'jwt_secret': 'secretpass',
            'max_iterations': 30,
        },
        'llm': {
            'input_cost_per_token': get_input_price_per_token(),
            'output_cost_per_token': get_output_price_per_token(),
            'log_completions': True,
            'log_completions_folder': '../logs/openhands', 
        },
        'sandbox':{
            'enable_gpu': has_gpu,
        }
    }

    # Define patterns for categorization
    patterns = {
        'core': [
            r'FILE_STORE',
            r'DATABASE',
            r'PORT',
            r'HOST'
        ], 
        'llm': [
            r'.*_API_BASE',
            r'.*_API_VERSION',
            r'.*MODEL',
            r'EMBEDDING',
            r'DEPLOYMENT',
            r'.*_SECRET',
            r'.*_KEY',
            r'.*_TOKEN'
        ]
    }

    for key, value in env_vars.items():
        # Check each pattern category
        for section, pattern_list in patterns.items():
            if any(re.match(pattern, key, re.IGNORECASE) for pattern in pattern_list):
                # Convert key to lowercase for consistency
                config_key = key.lower()

                # Standardize naming for common keys
                if '_API_KEY' in key:
                    config_key = 'api_key'
                elif '_API_BASE' in key or '_API_URL' in key:
                    config_key = 'base_url'
                elif '_API_VERSION' in key:
                    config_key = 'api_version'
                elif 'ORGANIZATION' in key:
                    continue  # Skip organization key

                config[section][config_key] = value
                break

    # Remove empty sections
    return {k: v for k, v in config.items() if v}

def setup_openhands_credential():
    """Convert an environment string to a TOML configuration."""
    try:
        with open("setup/env.sh", "r") as f:
            env_string = f.read()
        
        env_vars = parse_env_string(env_string)
        config = categorize_variables(env_vars)
        
        # Ensure directory exists
        output_path = Path("../workspace/config.toml")
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        # write to config.toml
        with open(output_path, "w") as f:
            f.write(toml.dumps(config))
            
        print(f'Set up OpenHands credentials in workspace/config.toml')
        return toml.dumps(config)  # Returns TOML as a string
        
    except Exception as e:
        print(f"Error setting up credentials: {str(e)}")
        raise

def load_system_prompt(prompt_path, **kwargs):
    with open(prompt_path, "r") as f:
        template = f.read()
    return template.format(**kwargs)
    


import pytest
from pydantic import ValidationError
from typing import Annotated, Literal
from typing_extensions import TypedDict
from langgraph.graph.message import add_messages
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import model
import scheduler as sched
import tool

from langgraph.store.memory import InMemoryStore

class State(TypedDict):
    # Messages have the type "list". The `add_messages` function
    # in the annotation defines how this state key should be updated
    # (in this case, it appends messages to the list, rather than overwriting them)
    messages: Annotated[list, add_messages]

state_instance: State = {
    "messages": []
}

store = InMemoryStore()
metadata_store = InMemoryStore()

store_write_tool = tool.NewExpPlanStoreWriteTool(store, metadata_store)
removal_plan_tool = tool.ArchiveExpPlanTool(store, metadata_store)

user_id = "admin"
application_context = "exp-sched" 
sched_namespace = (user_id, application_context)

# Necessary now:
user_input = "test question"
metadata_store.put(sched_namespace, "question", user_input)

sched.setup_sched(metadata_store, sched_namespace)

def test_valid_remove_plan():
    """
    Test that plan removal works given an existing valid plan_id. 
    """

    # Initialize and store the plan:
    valid_input = {
        "plan": {'hypothesis': 'AWS EC2 t3.medium instances are less CPU efficient compared to c5.large instances under a fixed compute-bound workload.', 'constant_vars': ['workload'], 'independent_vars': ['instance_type'], 'dependent_vars': ['execution_time'], 'controlled_experiment_setup_description': 'Create EC2 VM, run compute-bound task, measure execution time.', 'priority': 1, 'control_group': [{'instance_type': 't3.medium'}], 'experimental_group': [{'instance_type': 'c5.large'}]}
    }

    plan_id = store_write_tool.invoke(valid_input, state_instance)
    store_get_tool = tool.StoreGetTool(store)
    vals = store_get_tool.invoke({"plan_key": plan_id})

    print("The stored plan is structured as: ", vals)

    valid_input = {
        "plan_id": plan_id, 
    }

    # Store valid metadata:
    memory_id = str("supervisor_wrote_list")
    new_list = [plan_id]
    metadata_store.put(sched_namespace, memory_id, new_list)
    supervisor_wrote_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(supervisor_wrote_list) == 1

    memory_id = str("standby_exp_plan_list")
    new_list = [plan_id]
    metadata_store.put(sched_namespace, memory_id, new_list)
    standby_exp_plan_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(standby_exp_plan_list) == 1

    memory_id = str("supervisor_redo_partition_list")
    new_list = [{"plan_id": plan_id}]
    metadata_store.put(sched_namespace, memory_id, new_list)
    supervisor_redo_partition_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(supervisor_redo_partition_list) == 1

    # Remove the plan:
    return_string = removal_plan_tool.invoke(valid_input, state_instance)

    assert return_string == "Plan removal successful."

    # Check that the plan was removed from all metadata lists:
    memory_id = str("supervisor_wrote_list")
    wrote_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(wrote_list) == 0

    memory_id = str("standby_exp_plan_list")
    standby_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(standby_list) == 0

    memory_id = str("supervisor_redo_partition_list")
    redo_partition_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(redo_partition_list) == 0


def test_invalid_remove_plan():
    """
    Test that plan removal does not work and has no effect on other plans given a non-existing invalid plan_id. 
    """

    # Initialize and store the plan:
    valid_input = {
        "plan": {'hypothesis': 'AWS EC2 t3.medium instances are less CPU efficient compared to c5.large instances under a fixed compute-bound workload.', 'constant_vars': ['workload'], 'independent_vars': ['instance_type'], 'dependent_vars': ['execution_time'], 'controlled_experiment_setup_description': 'Create EC2 VM, run compute-bound task, measure execution time.', 'priority': 1, 'control_group': [{'instance_type': 't3.medium'}], 'experimental_group': [{'instance_type': 'c5.large'}]}
    }

    plan_id = store_write_tool.invoke(valid_input, state_instance)
    store_get_tool = tool.StoreGetTool(store)
    vals = store_get_tool.invoke({"plan_key": plan_id})

    print("The stored plan is structured as: ", vals)

    valid_input = {
        "plan_id": plan_id, 
    }

    # Store valid metadata:
    user_id = "admin"
    application_context = "exp-sched" 
    sched_namespace = (user_id, application_context)
    memory_id = str("supervisor_wrote_list")
    new_list = [plan_id]
    metadata_store.put(sched_namespace, memory_id, new_list)
    supervisor_wrote_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(supervisor_wrote_list) == 1

    memory_id = str("standby_exp_plan_list")
    new_list = [plan_id]
    metadata_store.put(sched_namespace, memory_id, new_list)
    standby_exp_plan_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(standby_exp_plan_list) == 1

    memory_id = str("supervisor_redo_partition_list")
    new_list = [{"plan_id": plan_id}]
    metadata_store.put(sched_namespace, memory_id, new_list)
    supervisor_redo_partition_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(supervisor_redo_partition_list) == 1

    # Attempt to Remove an invalid plan:
    invalid_input = {
        "plan_id": "invalid_plan_id", 
    }
    return_string = removal_plan_tool.invoke(invalid_input, state_instance)

    assert return_string == "The plan does not exist."

    # Check that the plan was removed from all metadata lists:
    memory_id = str("supervisor_wrote_list")
    wrote_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(wrote_list) == 1

    memory_id = str("standby_exp_plan_list")
    standby_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(standby_list) == 1

    memory_id = str("supervisor_redo_partition_list")
    redo_partition_list = metadata_store.get(sched_namespace, memory_id).dict()["value"]
    assert len(redo_partition_list) == 1

print(test_valid_remove_plan())
print(test_invalid_remove_plan())




# References used:
# https://langchain-ai.github.io/langgraph/concepts/human_in_the_loop
from nodes.base_node import BaseNode, NodeConfig
from langgraph.graph import StateGraph, START, END
from scheduler import SchedNode
from langgraph.types import interrupt, Command
from langchain_core.messages import HumanMessage, SystemMessage 

import os
import json

import settings
import model
import utils

class UserInput(BaseNode):

    def __init__(self, sched_node: SchedNode, config: NodeConfig, State, store, metadata_store, memory, tools: list):
        super().__init__(sched_node, config, State, store, metadata_store, memory, tools)  # Call parent class's __init__
        self.create_transition_objs()

    def create_transition_objs(self):
        # intro_message = "Here is the user's response:\n"
        self.node_config.transition_objs["done"] = lambda user_response: {
            "messages": user_response, 
            "next_agent": "user_input_router"
        }

    def transition_handle_func(self, state):
        """
        We will only forward to user_input_router. 
        """
        self.curie_logger.info("------------ Handle User Input ------------")

        items = self.store.search(self.plan_namespace)
        plans_list = [item.dict()["value"] for item in items]

        user_response = "Here are the previously proposed plans:\n" + str(plans_list) + "\n\nHere is the user's response: " + state["messages"][-1].content

        return self.node_config.transition_objs["done"](user_response)

    def create_subgraph(self):
        """ Creates a Node subgraph specific to UserInput. Override BaseNode implementation."""

        subgraph_builder = StateGraph(self.State)
        system_prompt_file = "dummy_system_prompt.txt"
        subgraph_node = self._create_model_response(system_prompt_file) 

        subgraph_builder.add_node(self.node_config.name, subgraph_node)
        subgraph_builder.add_edge(START, self.node_config.name)
        # subgraph_builder.add_edge(self.node_config.name, END)

        subgraph = subgraph_builder.compile(checkpointer=self.memory)
        os.makedirs("../../logs/misc") if not os.path.exists("../../logs/misc") else None
        # utils.save_langgraph_graph(subgraph, f"../../logs/misc/{self.node_config.name}_graph_image.png") 

        def call_subgraph(state: self.State) -> self.State: 
            response = subgraph.invoke({
                    "messages": state["messages"][-1]
                },
                {
                    # "recursion_limit": 20,
                    "configurable": {
                        "thread_id": f"{self.node_config.name}_graph_id"
                    }
                }
            )
            # ANSI escape codes for colors
            RED = "\033[91m"
            GREEN = "\033[92m"
            CYAN = "\033[96m"
            BOLD = "\033[1m"
            RESET = "\033[0m"

            # Custom input prompt
            user_input = input(f"{BOLD}{GREEN}✅ Do you approve of the architect's proposed plan? {RESET}"
                                f"{RED}❌ If not, why? {RESET}"
                                f"{CYAN}💬 Provide your response here: {RESET}")
            response = subgraph.invoke(Command(resume=user_input),
                {
                    # "recursion_limit": 20,
                    "configurable": {
                        "thread_id": f"{self.node_config.name}_graph_id"
                    }
                }
            )
            
            return {
                "messages": [
                    HumanMessage(content=response["messages"][-1].content, name=f"{self.node_config.name}_graph")
                ],
                "prev_agent": response.get("prev_agent", "scheduler"),
                "remaining_steps_display": state["remaining_steps"],
            }
        return call_subgraph

    def _create_model_response(self, system_prompt_file): 
        """Override BaseNode implementation."""   
        # FIXME: better way to get model names; from config?
        # FIXME: can move model name to model.py 
        def Node(state: self.State):
            if state["remaining_steps"] <= settings.CONCLUDER_BUFFER_STEPS:
                return {
                    "messages": [], 
                    "prev_agent": self.node_config.name,
                }

            response = interrupt("Do you approve of the architect's proposed plan?")

            self.curie_logger.info(f"<><><><><> {self.node_config.node_icon} {self.node_config.name.upper()} {self.node_config.node_icon} <><><><><>")

            self.curie_logger.info(f"Full response from user: {self.node_config.node_icon}: {response}")

            return {"messages": [response], "prev_agent": self.node_config.name}
        
        return Node

class UserInputRouter(BaseNode):

    def __init__(self, sched_node: SchedNode, config: NodeConfig, State, store, metadata_store, memory, tools: list):
        super().__init__(sched_node, config, State, store, metadata_store, memory, tools)  # Call parent class's __init__
        self.create_transition_objs()

    def create_transition_objs(self):
        intro_message = "You need to record down your decision using the user_router_wrote_list tool.\n"
        self.node_config.transition_objs["progress_not_recorded"] = lambda: {
            "messages": intro_message, 
            "next_agent": self.node_config.name
        }
        intro_message2 = "The user is satisfied with your proposed plan. Terminate and move on.\n"
        self.node_config.transition_objs["inform_architect_done"] = lambda: {
            "messages": intro_message2, 
            "next_agent": "supervisor",
            "is_user_input_done": True,
        }

        # intro_message = "The user is not satisfied with your proposed plan. Review your existing plans (via 'exp_plan_get'), then delete the plans (via 'exp_plan_remove'), then finally review the user's feedback below and propose a new plan that satisfies it:\n"
        intro_message3 = "The user is not satisfied with your previous proposed plan(s), which are attached for your reference. Review the user's feedback below and propose a new plan that satisfies it:\n"
        self.node_config.transition_objs["redo_architect"] = lambda router_log_message: {
            "messages": intro_message3 + router_log_message, 
            "next_agent": "supervisor"
        }

    def transition_handle_func(self, state):
        """
        Our router will essentially be called after the user input node. The router will analyze the user input, record its decision using the appropriate write tool. Here, we will analyze its decision, and decide what node to call next. Currently, there are only 2 possible choices:
            - If the user decides that the architect's plan is insufficient, we will go back to the architect node, and attach the router's analysis as context.
            - Otherwise, we will also return to architect, but let it know it is done.
        """
        self.curie_logger.info("------------ Handle User Input Router ------------")

        self.curie_logger.info("Checking user_router_wrote_list..")
        memory_id = str("user_router_wrote_list")
        wrote_list = self.metadata_store.get(self.sched_namespace, memory_id).dict()["value"]
        if len(wrote_list) == 0:
            return self.node_config.transition_objs["progress_not_recorded"]()
        if len(wrote_list) > 1:
            self.curie_logger.debug("WARNING: user input router wrote to user_router_wrote_list more than once..")

        # Reset wrote_list:
        self.metadata_store.put(self.sched_namespace, memory_id, [])

        if wrote_list[-1]["is_correct"]:
            return self.node_config.transition_objs["inform_architect_done"]()
        else:
            # Retrieve existing plans:
            items = self.store.search(self.plan_namespace)
            plans_list = [item.dict()["value"] for item in items]
            router_log_message = wrote_list[-1]["router_log_message"]

            feedback_msg = "Previous plan(s):\n" + str(plans_list) + "\n\nUser feedback:\n" + router_log_message

            # Delete all existing plans:
            for plan in plans_list:
                self.store.delete(self.plan_namespace, plan["plan_id"])
            # Reset architect wrote list:
            memory_id = str("supervisor_wrote_list")
            self.metadata_store.put(self.sched_namespace, memory_id, [])

            return self.node_config.transition_objs["redo_architect"](feedback_msg)


# NOTE: remember to export env variables first for the LLM model. 
# Set up the state
from langgraph.graph import MessagesState, START
import os
# Set up the tool
# We will have one real tool - a search tool
# We'll also have one "fake" tool - a "ask_human" tool
# Here we define any ACTUAL tools
from langchain_core.tools import tool
from langchain_community.chat_models import ChatLiteLLM
from langgraph.prebuilt import ToolNode
from IPython.display import Image, display
from langgraph.types import Command, interrupt


@tool
def search(query: str):
    """Call to surf the web."""
    # This is a placeholder for the actual implementation
    # Don't let the LLM know this though 😊
    return f"I looked up: {query}. Result: It's sunny in San Francisco, but you better look out if you're a Gemini 😈."


tools = [search]
tool_node = ToolNode(tools)

# Set up the model

model = ChatLiteLLM(model=os.environ.get("MODEL"))

from pydantic import BaseModel


# We are going "bind" all tools to the model
# We have the ACTUAL tools from above, but we also need a mock tool to ask a human
# Since `bind_tools` takes in tools but also just tool definitions,
# We can define a tool definition for `ask_human`
class AskHuman(BaseModel):
    """Ask the human a question"""

    question: str


model = model.bind_tools(tools + [AskHuman])

# Define nodes and conditional edges


# Define the function that determines whether to continue or not
def should_continue(state):
    messages = state["messages"]
    last_message = messages[-1]
    # If there is no function call, then we finish
    if not last_message.tool_calls:
        return END
    # If tool call is asking Human, we return that node
    # You could also add logic here to let some system know that there's something that requires Human input
    # For example, send a slack message, etc
    elif last_message.tool_calls[0]["name"] == "AskHuman":
        return "ask_human"
    # Otherwise if there is, we continue
    else:
        return "action"


# Define the function that calls the model
def call_model(state):
    messages = state["messages"]
    response = model.invoke(messages)
    # We return a list, because this will get added to the existing list
    return {"messages": [response]}


# We define a fake node to ask the human
def ask_human(state):
    tool_call_id = state["messages"][-1].tool_calls[0]["id"]
    location = interrupt("Please provide your location:")
    tool_message = [{"tool_call_id": tool_call_id, "type": "tool", "content": location}]
    return {"messages": tool_message}


# Build the graph

from langgraph.graph import END, StateGraph

# Define a new graph
workflow = StateGraph(MessagesState)

# Define the three nodes we will cycle between
workflow.add_node("agent", call_model)
workflow.add_node("action", tool_node)
workflow.add_node("ask_human", ask_human)

# Set the entrypoint as `agent`
# This means that this node is the first one called
workflow.add_edge(START, "agent")

# We now add a conditional edge
workflow.add_conditional_edges(
    # First, we define the start node. We use `agent`.
    # This means these are the edges taken after the `agent` node is called.
    "agent",
    # Next, we pass in the function that will determine which node is called next.
    should_continue,
)

# We now add a normal edge from `tools` to `agent`.
# This means that after `tools` is called, `agent` node is called next.
workflow.add_edge("action", "agent")

# After we get back the human response, we go back to the agent
workflow.add_edge("ask_human", "agent")

# Set up memory
from langgraph.checkpoint.memory import MemorySaver

memory = MemorySaver()

# Finally, we compile it!
# This compiles it into a LangChain Runnable,
# meaning you can use it as you would any other runnable
# We add a breakpoint BEFORE the `ask_human` node so it never executes
app = workflow.compile(checkpointer=memory)

display(Image(app.get_graph().draw_mermaid_png()))


config = {"configurable": {"thread_id": "2"}}
for event in app.stream(
    {
        "messages": [
            (
                "user",
                "Use the search tool to ask the user where they are, then look up the weather there",
            )
        ]
    },
    config,
    stream_mode="values",
):
    event["messages"][-1].pretty_print()

# app.get_state(config).next
input_1 = input()

for event in app.stream(Command(resume=input_1), config, stream_mode="values"):
    event["messages"][-1].pretty_print()


import tool

prompt='''
The starter file can be found under "/starter_file/large_language_monkeys".
Note: there is no need to modify the source code of any files outside "/starter_file/large_language_monkeys". Though you may choose to read the source code just to enhance your understanding of its inner workings.

Setup the python environment for llmonk following the large_language_monkeys/README.md. 

Clean up the files under ./logs. 

OpenAI Azure credentials are available under large_language_monkeys/env.sh

You can verify the success of the setup by running the following command: python llmonk/generate/gsm8k.py

Provide a reproducible experimental workflow named experimental_workflow.sh, which contains the steps to setup the environment.
'''

prompt='''
The starter file can be found under "/workspace".
write a quick sort function in a new file called 'sort.py' under the starter file directory
'''
res = tool.codeagent_openhands(prompt)
print("===========================================")
print(res) 


from nodes.base_node import BaseNode, NodeConfig
from langgraph.graph import END
from langchain_core.messages import SystemMessage
from scheduler import SchedNode
import model
import settings

class Architect(BaseNode):

    def __init__(self, sched_node: SchedNode, config: NodeConfig, State, store, metadata_store, memory, tools: list):
        super().__init__(sched_node, config, State, store, metadata_store, memory, tools)  # Call parent class's __init__
        self.create_transition_objs()

    def create_transition_objs(self):
        # TODO: simplify this init message.
        intro_message = "This is the question to answer, analyze the user's question and mentioned or referenced files (e.g., .txt, .md, .pdf). \
            Use `read_file_contents` for non-PDF files and `query_pdf` for PDFs to extract relevant information that informs your experimental design. \
            Make sure to formulate it in terms of an experimental plan(s) using the 'write_new_exp_plan' tool:\n"
        self.node_config.transition_objs["no_plan"] = lambda: {"messages": intro_message + self.sched_node.get_question(), "next_agent": "supervisor", "prev_agent": "supervisor"}

        self.node_config.transition_objs["get_user_input_first"] = lambda: {
            "messages": "[User input requested]", 
            "next_agent": "user_input"
        }

        self.node_config.transition_objs["is_terminate"] = lambda: {"next_agent": END}

        self.node_config.transition_objs["control_has_work"] = lambda assignment_messages: {
            "control_work": {"messages": assignment_messages, "next_agent": "control_worker"},
            "experimental_work": {"messages": [], "next_agent": "control_worker"}
        }

        self.node_config.transition_objs["experimental_has_work"] = lambda assignment_messages: {
            "control_work": {"messages": assignment_messages, "next_agent": "control_worker"},
            "experimental_work": {"messages": self.sched_node.assign_worker("experimental"), "next_agent": "worker"}
        }

    def transition_handle_func(self, state):
        """
        All of the plans that were edited/written by the supervisor will be scheduled for execution now or added to the queue. 
        If the plan existed before, we check if the experimental groups have been modified: 
            if yes, we will preempt the existing group (TODO: not implemented yet. this requires that we for instance also record the old plan to know if it has changed..) from the current worker, 
            or modify within the queue. 
        If the plan is new, 
            we simply add to worker if some are idle, 
            or add to queue.
        """
        self.curie_logger.info("------------ Handle Supervisor ------------")

        # Zero, if no plan exists at all, we need to re-prompt the architect to force it to create a plan:
        if self.sched_node.is_no_plan_exists():
            return self.node_config.transition_objs["no_plan"]()
        # print("state['is_user_input_done']: ", state["is_user_input_done"])
        if not state["is_user_input_done"]:
            return self.node_config.transition_objs["get_user_input_first"]()

        # First, check for exp termination condition:
        is_terminate = self.sched_node.check_exp_termination_condition()
        if is_terminate:
            return self.node_config.transition_objs["is_terminate"]()

        # Second, for control groups that are done, move their experimental groups (if any exist) to the worker queue:
        self.curie_logger.debug("Checking control group done..")
        memory_id = str("standby_exp_plan_list")
        standby_exp_plan_list = self.metadata_store.get(self.sched_node.sched_namespace, memory_id).dict()["value"]
        for plan_id in standby_exp_plan_list[:]: # iterate over a copy of standby_exp_plan_list
            self.sched_node.update_queues(plan_id, assert_new_control=True)

        # Third, for plans that were written/modified by the supervisor, add them to the correct queue:
        self.curie_logger.debug("Checking supervisor wrote list..")
        memory_id = str("supervisor_wrote_list")
        supervisor_wrote_list = self.metadata_store.get(self.sched_node.sched_namespace, memory_id).dict()["value"]
        # Create a new /workspace dir for each new plan_id, and other related inits for a new plan:
        self.sched_node.init_new_plan(supervisor_wrote_list)

        # Update queues:
        for _ in range(len(supervisor_wrote_list)):
            # NOTE: currently, we ignore plan groups that are already executing in the worker.. 
            plan_id = supervisor_wrote_list.pop(0)
            self.sched_node.update_queues(plan_id) 
        # Reset supervisor_wrote_list:
        assert len(supervisor_wrote_list) == 0
        self.metadata_store.put(self.sched_node.sched_namespace, memory_id, supervisor_wrote_list)

        # Fourth, for partition's that need to be redone, add them to the correct queue: the difference here is that an explicit error feedback is provided by the supervisor.
        self.curie_logger.info("Checking supervisor redo partition list..")
        memory_id = str("supervisor_redo_partition_list")
        supervisor_redo_partition_list = self.metadata_store.get(self.sched_node.sched_namespace, memory_id).dict()["value"]
        for _ in range(len(supervisor_redo_partition_list)):
            redo_details = supervisor_redo_partition_list.pop(0)
            self.sched_node.update_queues(redo_details["plan_id"], redo_details=redo_details)
        # Reset supervisor_redo_partition_list:
        assert len(supervisor_redo_partition_list) == 0
        self.metadata_store.put(self.sched_node.sched_namespace, memory_id, supervisor_redo_partition_list)

        # Fifth, Assign to control and experimental workers if there are idle workers: according to priority
        # TODO: currently since we have NOT implemented ASYNC, we will only run 1 worker at a time (i.e., either control or normal worker). Note that we also only have 1 control and 1 normal worker only. Since there is no async I did not implement parallelism yet. 
        assignment_messages = self.sched_node.assign_worker("control")
        if not assignment_messages: # if there is no control group to run, we will run the experimental group of some partition
            return self.node_config.transition_objs["experimental_has_work"](assignment_messages)
        else: # if there exist a control group to run, we do not run the experimental group of any partition
            return self.node_config.transition_objs["control_has_work"](assignment_messages)

        # Original version:  
        # return {
        #     "control_work": {"messages": self.assign_worker("control"), "next_agent": "control_worker"},
        #     "experimental_work": {"messages": self.assign_worker("experimental"), "next_agent": "worker"}
        # }
 
    def _create_model_response(self, system_prompt_file):     
        def Node(state: self.State):
            if state["remaining_steps"] <= settings.CONCLUDER_BUFFER_STEPS:
                return {
                    "messages": [], 
                    "prev_agent": self.node_config.name,
                }
                
            # Read from prompt file:
            with open(system_prompt_file, "r") as file:
                system_prompt = file.read()

            system_message = SystemMessage(
                content=system_prompt,
            )  
 
            messages = state["messages"]

            # Ensure the system prompt is included at the start of the conversation
            if not any(isinstance(msg, SystemMessage) for msg in messages):
                messages.insert(0, system_message)

            self.curie_logger.debug(f"Messages TO {self.node_config.name.upper()} {self.node_config.node_icon}: {messages}")
            
            response = model.query_model_safe(messages, tools=self.tools)

            self.curie_logger.info(f"<><><><><> {self.node_config.node_icon} {self.node_config.name.upper()} {self.node_config.node_icon} <><><><><>")

            if response.tool_calls:
                self.curie_logger.info(f"Tool calls: {response.tool_calls[0]['name']}")
                if 'prompt' in response.tool_calls[0]['args']:
                    self.curie_logger.info(f"Message received: {response.tool_calls[0]['args']['prompt']}")
                elif 'verifier_log_message' in response.tool_calls[0]['args']:
                    self.curie_logger.info(f"Message: {response.tool_calls[0]['args']['verifier_log_message']}")
                else:
                    self.curie_logger.info(f"Message: {response.tool_calls[0]['args']}")
            
            concise_msg = response.content.split('\n\n')[0] 
            self.curie_logger.info(f'Concise response: {concise_msg}')

            self.curie_logger.debug(f"Full response from {self.node_config.name.upper()} {self.node_config.node_icon}: {response}")

            return {"messages": [response], "prev_agent": self.node_config.name}
        
        return Node



import os
import sys
import json
import traceback
from typing import Annotated, Literal
from typing_extensions import TypedDict 

from langgraph.graph import StateGraph, START, END
from langgraph.graph.message import add_messages
from langgraph.types import interrupt, Command
from langgraph.store.memory import InMemoryStore
from langgraph.checkpoint.memory import MemorySaver
from langchain_core.messages import HumanMessage, SystemMessage
from langgraph.managed.is_last_step import RemainingSteps

# Local module imports
import model
import tool
import settings
import scheduler as sched

from logger import init_logger
from model import setup_model_logging, report_cost_stats
from tool import setup_tool_logging
from nodes.exec_validator import setup_exec_validator_logging

from nodes.architect import Architect
from nodes.technician import Technician
from nodes.base_node import NodeConfig
from nodes.llm_validator import LLMValidator
from nodes.patcher import Patcher
from nodes.analyzer import Analyzer
from nodes.concluder import Concluder
from nodes.user_input import UserInput, UserInputRouter
from reporter import generate_report

if len(sys.argv) < 2:
    print("Usage: python script.py <config_file>")
    sys.exit(1)

config_filename = sys.argv[1]

# Read config file
with open(config_filename, 'r') as file:
    config = json.load(file)
    exp_plan_filename = f"../{config['exp_plan_filename']}"
    log_filename = f"../{config['log_filename']}"
    log_file = open(log_filename, 'w')
    
    curie_logger = init_logger(log_filename)
    setup_model_logging(log_filename)
    setup_exec_validator_logging(log_filename)
    setup_tool_logging(log_filename)

class State(TypedDict):
    messages: Annotated[list, add_messages]
    prev_agent: Literal[*settings.AGENT_LIST]
    next_agent: Literal[*settings.AGENT_LIST]
    is_terminate: bool 
    is_user_input_done: bool
    remaining_steps: RemainingSteps
    remaining_steps_display: int # remaining_steps cannot be seen in event.values since it is an Annotated value managed by RemainingStepsManager https://github.com/langchain-ai/langgraph/blob/main/libs/langgraph/langgraph/managed/is_last_step.py

class AllNodes():
    def __init__(self, log_filename: str, config_filename: str, state: State, store, metadata_store, memory):
        self.nodes = {}
        self.State = State
        self.log_filename = log_filename
        self.store = store
        self.metadata_store = metadata_store
        self.memory = memory
        self.config_filename = config_filename
        self.instantiate_nodes()
        self.instantiate_subgraphs()
    
    def instantiate_nodes(self):
        with open(self.config_filename, 'r') as file:
            config_dict = json.load(file)

        # Create scheduler node:
        self.sched_node = self.create_sched_node(config_dict) # always create sched node first

        # Create other nodes, passing in self.sched_node as a param:
        self.architect = self.create_architect_node()
        worker, control_worker = self.create_worker_nodes()
        self.workers = [worker]
        self.control_workers = [control_worker]
        self.validators = self.create_validators() # list of validators
        self.user_input_nodes = self.create_user_input_nodes()

        # Create sched tool, passing in other agent's transition funcs as a dict
        config_dict["transition_funcs"] = {
            "supervisor": lambda state: self.architect.transition_handle_func(state),
            "worker": lambda: self.workers[0].transition_handle_func(),
            "control_worker": lambda: self.control_workers[0].transition_handle_func(),
            "llm_verifier": lambda: self.validators[0].transition_handle_func(),
            "patch_verifier": lambda: self.validators[1].transition_handle_func(),
            "analyzer": lambda: self.validators[2].transition_handle_func(),
            "concluder": lambda state: self.validators[3].transition_handle_func(state),
            "user_input": lambda state: self.user_input_nodes[0].transition_handle_func(state),
            "user_input_router": lambda state: self.user_input_nodes[1].transition_handle_func(state),
        }
        self.sched_tool = sched.SchedTool(self.store, self.metadata_store, config_dict)

    def instantiate_subgraphs(self):
        self.sched_subgraph = self.sched_node.create_SchedNode_subgraph(self.sched_tool)
        self.architect_subgraph = self.architect.create_subgraph()
        self.worker_subgraph = self.workers[0].create_subgraph()
        self.control_worker_subgraph = self.control_workers[0].create_subgraph()
        self.validator_subgraphs = [validator.create_subgraph() for validator in self.validators]
        self.user_input_subgraphs = [user_input.create_subgraph() for user_input in self.user_input_nodes]
    
    def get_sched_subgraph(self):
        return self.sched_subgraph
    
    def get_architect_subgraph(self):
        return self.architect_subgraph
    
    def get_worker_subgraphs(self):
        return self.worker_subgraph, self.control_worker_subgraph
    
    def get_validator_subgraphs(self):
        return self.validator_subgraphs

    def get_user_input_subgraphs(self):
        return self.user_input_subgraphs

    def get_architect_node(self):
        return self.architect

    def get_worker_node(self):
        return self.workers[0]
    
    def get_control_worker_node(self):
        return self.control_workers[0]
    
    def get_validator_nodes(self):
        return self.validators

    def get_user_input_nodes(self):
        return self.user_input_nodes

    def create_sched_node(self, config_dict):
        return sched.SchedNode(self.store, self.metadata_store, self.State, config_dict)

    def create_architect_node(self):
        # Customizable node config 
        node_config = NodeConfig(
            name="supervisor",
            node_icon="👑",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="supervisor_system_prompt_filename",
            default_system_prompt_filename="prompts/exp-supervisor.txt"
        )
        # Customizable tools
        store_write_tool = tool.NewExpPlanStoreWriteTool(self.store, self.metadata_store)
        redo_write_tool = tool.RedoExpPartitionTool(self.store, self.metadata_store)
        store_get_tool = tool.StoreGetTool(self.store)
        edit_priority_tool = tool.EditExpPriorityTool(self.store, self.metadata_store)
        with open(self.config_filename, 'r') as file:
            config_dict = json.load(file) 
        pdf_query_tool = tool.QueryPDFTool(config_dict)

        tools = [store_write_tool, edit_priority_tool, redo_write_tool, store_get_tool, \
                 tool.read_file_contents, pdf_query_tool]

        return Architect(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

    def create_worker_nodes(self):
        # Create common tools:
        # Customizable tools
        store_write_tool = tool.ExpPlanCompletedWriteTool(self.store, self.metadata_store)
        store_get_tool = tool.StoreGetTool(self.store)
        with open(self.config_filename, 'r') as file:
            config_dict = json.load(file) 
        codeagent_openhands = tool.CodeAgentTool(config_dict)
        pdf_query_tool = tool.QueryPDFTool(config_dict)
        tools = [codeagent_openhands, pdf_query_tool, tool.execute_shell_command, store_write_tool, store_get_tool]

        # Create 1 worker: 
        # Customizable node config 
        worker_names = settings.list_worker_names()
        assert len(worker_names) == 1
        node_config = NodeConfig(
            name=worker_names[0],
            node_icon="👷",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="worker_system_prompt_filename",
            default_system_prompt_filename="prompts/exp-worker.txt"
        )

        worker = Technician(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        # Create 1 control worker: 
        # Customizable node config 
        worker_names = settings.list_control_worker_names()
        assert len(worker_names) == 1
        node_config = NodeConfig(
            name=worker_names[0],
            node_icon="👷",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="control_worker_system_prompt_filename",
            default_system_prompt_filename="prompts/controlled-worker.txt"
        )

        control_worker = Technician(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        return worker, control_worker

    def create_validators(self):
        # Create LLM validator: 
        # Customizable node config 
        node_config = NodeConfig(
            name="llm_verifier",
            node_icon="✅",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="llm_verifier_system_prompt_filename",
            default_system_prompt_filename="prompts/llm-verifier.txt"
        )

        # Customizable tools
        verifier_write_tool = tool.LLMVerifierWriteTool(self.store, self.metadata_store)
        store_get_tool = tool.StoreGetTool(self.store)
        tools = [tool.execute_shell_command, store_get_tool, verifier_write_tool]
        
        llm_validator = LLMValidator(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        # Create Patcher: 
        # Customizable node config 
        node_config = NodeConfig(
            name="patch_verifier",
            node_icon="✅",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="patcher_system_prompt_filename",
            default_system_prompt_filename="prompts/exp-patcher.txt"
        )

        # Customizable tools
        patcher_record_tool = tool.PatchVerifierWriteTool(self.store, self.metadata_store)
        with open(self.config_filename, 'r') as file:
            config_dict = json.load(file) 
        patch_agent_tool = tool.PatcherAgentTool(config_dict)
        store_get_tool = tool.StoreGetTool(self.store)
        tools = [patch_agent_tool, tool.execute_shell_command, patcher_record_tool, store_get_tool] 
        
        patcher = Patcher(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        # Create Analyer: 
        # Customizable node config 
        node_config = NodeConfig(
            name="analyzer",
            node_icon="✅",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="analyzer_system_prompt_filename",
            default_system_prompt_filename="prompts/exp-analyzer.txt"
        )

        # Customizable tools
        patcher_record_tool = tool.AnalyzerWriteTool(self.store, self.metadata_store)
        store_get_tool = tool.StoreGetTool(self.store)
        tools = [tool.read_file_contents, patcher_record_tool, store_get_tool]
        
        analyzer = Analyzer(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        # Create Concluder: 
        # Customizable node config 
        node_config = NodeConfig(
            name="concluder",
            node_icon="✅",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="concluder_system_prompt_filename",
            default_system_prompt_filename="prompts/exp-concluder.txt"
        )

        # Customizable tools
        patcher_record_tool = tool.ConcluderWriteTool(self.store, self.metadata_store)
        store_get_tool = tool.StoreGetTool(self.store)
        tools = [tool.read_file_contents, patcher_record_tool, store_get_tool] # Only tool is code execution for now
        
        concluder = Concluder(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        return [llm_validator, patcher, analyzer, concluder]

    def create_user_input_nodes(self):
        # Customizable node config 
        node_config = NodeConfig(
            name="user_input",
            node_icon="💬",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="dummy_prompt",
            default_system_prompt_filename="prompts/exp-dummy-prompt.txt"
        )
        tools = []
        user_input = UserInput(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        # Customizable node config
        node_config = NodeConfig(
            name="user_input_router",
            node_icon="🔀",
            log_filename=self.log_filename, 
            config_filename=self.config_filename,
            system_prompt_key="user_input_router_system_prompt_filename",
            default_system_prompt_filename="prompts/exp-user-input-router.txt"
        )
        # Customizable tools
        router_write_tool = tool.UserInputRouterWriteTool(self.store, self.metadata_store)
        tools = [router_write_tool]
        user_input_router = UserInputRouter(self.sched_node, node_config, self.State, self.store, self.metadata_store, self.memory, tools)

        return [user_input, user_input_router]

    def get_all_nodes(self):
        return self.nodes

def setup_logging(log_filename: str):
    """
    Configure logging to redirect stdout and stderr to a log file.
    
    Args:
        log_filename (str): Path to the log file
    """
    log_file = open(log_filename, 'w')
    sys.stdout = log_file
    sys.stderr = log_file

def create_graph_stores():
    """
    Create stores and memory for the graph.
    
    Returns:
        tuple: Stores and memory objects
    """
    store = InMemoryStore()
    metadata_store = InMemoryStore()
    memory = MemorySaver()
    return store, metadata_store, memory

def build_graph(State, config_filename):
    """
    Build the complete LangGraph workflow.
    
    Args:
        State: Graph state type
        config_filename: Path to configuration file
    
    Returns:
        Compiled graph
    """
    # Read configuration
    with open(config_filename, 'r') as file:
        config = json.load(file)
    
    # Create stores
    store, metadata_store, memory = create_graph_stores()
    
    # Create graph builder
    graph_builder = StateGraph(State)

    all_nodes = AllNodes(log_filename, config_filename, State, store, metadata_store, memory)

    # Add scheduler node
    sched_subgraph = all_nodes.get_sched_subgraph()
    graph_builder.add_node("scheduler", sched_subgraph)
    
    # Add supervisor node
    supervisor_graph = all_nodes.get_architect_subgraph()
    supervisor_name = all_nodes.get_architect_node().get_name()
    graph_builder.add_node(supervisor_name, supervisor_graph)
    
    # Add worker nodes
    experimental_worker, control_worker = all_nodes.get_worker_subgraphs()
    experimental_worker_name = all_nodes.get_worker_node().get_name()
    control_worker_name = all_nodes.get_control_worker_node().get_name()

    graph_builder.add_node(experimental_worker_name, experimental_worker)
    graph_builder.add_node(control_worker_name, control_worker)
    
    # Add verification nodes
    verification_subgraphs = all_nodes.get_validator_subgraphs()
    verification_nodes = all_nodes.get_validator_nodes()
    for index, node in enumerate(verification_nodes):
        graph_builder.add_node(node.get_name(), verification_subgraphs[index])
    
    # Add user input nodes
    user_input_subgraphs = all_nodes.get_user_input_subgraphs()
    user_input_nodes = all_nodes.get_user_input_nodes()
    for index, node in enumerate(user_input_nodes):
        graph_builder.add_node(node.get_name(), user_input_subgraphs[index])
    
    # Add graph edges
    graph_builder.add_edge(START, supervisor_name)
    graph_builder.add_edge(supervisor_name, "scheduler")
    # graph_builder.add_conditional_edges("supervisor", router, ["scheduler", END])
    graph_builder.add_edge(experimental_worker_name, "scheduler")
    graph_builder.add_edge(control_worker_name, "scheduler")
    
    for _, node in enumerate(verification_nodes):
        graph_builder.add_edge(node.get_name(), "scheduler")

    for _, node in enumerate(user_input_nodes):
        graph_builder.add_edge(node.get_name(), "scheduler")
    
    graph_builder.add_conditional_edges("scheduler", lambda state: state["next_agent"])
    
    # Compile and visualize graph
    graph = graph_builder.compile(checkpointer=memory)
    # utils.save_langgraph_graph(graph, "../logs/misc/overall_graph_image.png")
    
    return graph, metadata_store, config

def get_question(question_file_path: str) -> str:
    """
    Read question from a file.

    Args:
        question_file_path (str): Path to the question file

    Returns:
        str: Question text
    """

    with open(question_file_path, "r") as question_file:
        question = question_file.read().strip() 
    return True, question

def validate_question(question: str) -> bool:
    with open('prompts/parse-input.txt', 'r') as file:
        parse_input_prompt = file.read().strip()
 
        # validate question, if it's feasible to answer through experimentation
        # if not just return the answer via LLM call,  and prompt the user to input a researchß question
        messages = [SystemMessage(content=parse_input_prompt),
                HumanMessage(content=question)]

        response = model.query_model_safe(messages)
        try:
            response_json = json.loads(response.content)
            valid = response_json["valid"]
            response = response_json["response"] if response_json["response"] else question
        except json.JSONDecodeError:
            response_json = {"valid": True, "response": None}
            valid = True
            response = ''
        return valid, response

def stream_graph_updates(graph, user_input: str, config: dict):
    """
    Stream graph updates during workflow execution.

    Args:
        graph: Compiled LangGraph workflow
        user_input (str): User's input question
    """
    max_global_steps = config.get("max_global_steps", 50)
    max_global_steps += settings.CONCLUDER_BUFFER_STEPS
    is_user_input_done = not config.get("is_user_interrupt_allowed", False) # true == no interrupt (at the architect plan design stage)
    # Prior to user-input interrupt:
    for event in graph.stream(
        {"messages": [("user", user_input)], "is_terminate": False, "is_user_input_done": is_user_input_done}, 
        {"recursion_limit": max_global_steps, "configurable": {"thread_id": "main_graph_id"}}
    ):
        print_graph_updates(event, max_global_steps)
    
    # # Resume with user input everytime we are interrupted, until the recursion limit is hit:
    # while True:
    #     user_input = input("Enter your response: ")
    #     for event in graph.stream(
    #         Command(resume=user_input), 
    #         {"recursion_limit": max_global_steps, "configurable": {"thread_id": "main_graph_id"}}
    #     ):
    #         print_graph_updates(event, max_global_steps)

def print_graph_updates(event, max_global_steps):
    event_vals = list(event.values())
    step = max_global_steps - event_vals[0]["remaining_steps_display"] # if there are multiple event values, we believe they will have the same remaining steps (only possible in parallel execution?)..
    curie_logger.info(f"============================ Global Step {step} ============================")    
    curie_logger.debug(f"Event: {event}")
    for value in event.values():
        curie_logger.info(f"Event value: {value['messages'][-1].content}")

def report_all_logs(config_filename: str, config: dict):
    exp_plan_dirname = config['log_filename'].split("/")[:-1]
    exp_plan_filename = '/' + '/'.join(exp_plan_dirname) + '/' + os.path.basename(config['exp_plan_filename']).replace('.txt', '.json')
    try: 
        plans = []
        with open(exp_plan_filename, 'r') as file:
            workspace_dir_list = []
            for line in file.readlines():
                if line == '\n':
                    continue
                plan = json.loads(line) 
                plans.append(plan)
                workspace_dir = plan['workspace_dir'].replace('/', '', 1)
                workspace_dir_list.append(workspace_dir) 
        # if config['report'] == True:
        report_filename, result_filename = generate_report(config, plans)
        curie_logger.info(f"📝 Experiment report saved to {report_filename[1:]}")
        curie_logger.info(f"📊 Experiment results saved to {result_filename[1:]}")
    
        curie_logger.info(f"📋 Raw experiment plan an be found in {exp_plan_filename.replace('/', '', 1)}")
        curie_logger.info(f"📁 Workspace is located at {workspace_dir_list}.")
    except Exception as e:
        curie_logger.error(f"⚠️ Failed to read experiment plan: {exp_plan_filename}. Error: {e}") 
    
    curie_logger.info("=================== Raw Curie Experiment Logs ==================")
    curie_logger.info(f"📋 Experiment plan can be found in {config_filename.replace('/', '', 1)}")
    curie_logger.info(f"📓 Experiment config file can be found in {config_filename.replace('/', '', 1)}")
    curie_logger.info(f"📒 Experiment loggings can be found in {config['log_filename']}")
    report_cost_stats()
    curie_logger.info("🎉 Experiment completed successfully!")

def main():
    """
    Main execution function for the LangGraph workflow.
    """
    if len(sys.argv) < 2:
        curie_logger.error("Usage: python script.py <config_file>")
        sys.exit(1)

    config_filename = sys.argv[1]

    try:
        # Build graph
        graph, metadata_store, config = build_graph(State, config_filename)

        # Read question from file
        exp_plan_filename = f"../{config['exp_plan_filename']}"
        valid, user_input = get_question(exp_plan_filename) 
        # if not valid:
        #     curie_logger.error(f"⚠️ Invalid question. Please input a valid research question.\n{user_input}")
        #     sys.exit(0)
        sched_namespace = ("admin", "exp-sched")
        metadata_store.put(sched_namespace, "question", user_input)

        # Stream graph updates
        stream_graph_updates(graph, user_input, config)
        report_all_logs(config_filename, config)

    except Exception as e:
        curie_logger.error(f"Execution error: {e}")
        traceback.print_exc()
        sys.exit(1)

if __name__ == "__main__":
    main()



from langchain_core.tools import tool
from typing import Annotated, List
from langgraph.store.memory import InMemoryStore
from langchain_core.messages import HumanMessage
from langgraph.prebuilt import InjectedStore
from langgraph.prebuilt import InjectedState
import uuid
from typing import Optional, Type, Dict

from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field

from collections import deque, defaultdict
import re

import formatter
import settings
import worker_agent

import scheduler as sched

import pytest
from pydantic import ValidationError
from typing import Annotated, Literal
from typing_extensions import TypedDict
from langgraph.graph.message import add_messages

import tool
from langgraph.store.memory import InMemoryStore

class State(TypedDict):
    # Messages have the type "list". The `add_messages` function
    # in the annotation defines how this state key should be updated
    # (in this case, it appends messages to the list, rather than overwriting them)
    messages: Annotated[list, add_messages]
    prev_agent: Literal[*settings.AGENT_LIST] # The agent that called the tool (we are using this mainly for the standard exp-agent interface functions, such as write, get..., since these are agent->tool interactions. Note that Sched tool does not need this, since we will always do a worker -> sched -> supervisor, or supervisor -> sched -> worker transitions) # To view prev_agent within tools, we need: https://langchain-ai.github.io/langgraph/how-tos/pass-run-time-values-to-tools/#pass-graph-state-to-tools
    next_agent: Literal[*settings.AGENT_LIST]

state: State = {
    "messages": [],
    "prev_agent": "supervisor",
    # "next_agent": "sched"
}

store = InMemoryStore()
metadata_store = InMemoryStore()

sched.setup_sched(metadata_store)

sched_tool = sched.SchedTool(store, metadata_store)
response = sched_tool.invoke({"state":state}) # response is guaranteed to be a dict
print(response)


from langchain_community.chat_models import ChatLiteLLM
from langchain_text_splitters import RecursiveCharacterTextSplitter
import tiktoken
import time
import os
from openai import BadRequestError
from typing import List, Dict, Any
from langchain_core.messages import BaseMessage, HumanMessage, SystemMessage
# import anthropic

import utils

from logger import init_logger
def setup_model_logging(log_filename: str):
    global curie_logger 
    curie_logger = init_logger(log_filename)

def report_cost_stats():
    """Print accumulated token usage and costs."""
    stats = TokenCounter.get_accumulated_stats()
    curie_logger.info(f"💸 Accumulated cost for input tokens: ${stats['costs']['input']:.4f}")
    curie_logger.info(f"💸 Accumulated cost for output tokens: ${stats['costs']['output']:.4f}")
    curie_logger.info(f"💸 Accumulated cost for tool usage: ${stats['costs']['tool_cost']:.4f}")

class TokenCounter:
    # Pricing per 1k tokens (as of March 2024)
    PRICE_PER_1K_TOKENS = utils.get_all_price_per_1k_tokens()

    # Class-level variables to track accumulated usage across all instances
    _accumulated_tokens = {"input": 0, "output": 0}
    # Can be more fine-grained once we introduce more tools
    _accumulated_cost = {"input": 0.0, "output": 0.0, "tool_cost": 0.0}

    def __init__(self):
        # Strip provider prefix if present (e.g., "openai/gpt-4" -> "gpt-4")
        self.model_name = utils.get_model_name()
        
        # try:
        #     self.encoding = tiktoken.encoding_for_model(self.model_name)
        # except KeyError:
        #     # Fall back to cl100k_base for models not in tiktoken
        #     self.encoding = tiktoken.get_encoding("cl100k_base")

    @classmethod
    def get_accumulated_stats(cls) -> Dict[str, Dict[str, float]]:
        """Get accumulated token usage and costs across all instances."""
        return {
            "tokens": dict(cls._accumulated_tokens),
            "costs": dict(cls._accumulated_cost),
            "total_cost": sum(cls._accumulated_cost.values())
        }
    
    def count_output_tokens(self, string: str) -> int:
        """
        MetaGPT anthropic client token counter does not work for anthropic>=0.39.0: https://github.com/geekan/MetaGPT/blob/main/metagpt/utils/token_counter.py#L479C1-L480C1
        Use simple tokenizer instead, since that is what langchain_aws is doing: 
            - https://github.com/langchain-ai/langchain-aws/commit/6355b0ff44c92b594ab8c3a5c50ac726904d716d
            - https://github.com/langchain-ai/langchain-aws/issues/314
            - https://python.langchain.com/api_reference/_modules/langchain_core/language_models/base.html#BaseLanguageModel

        Returns the number of tokens in a text string.

        Args:
            string (str): The text string.

        Returns:
            int: The number of tokens in the text string.
        """
        # if "claude" in self.model_name:
        #     vo = anthropic.Client()
        #     num_tokens = vo.count_tokens(string)
        #     return num_tokens
        try:
            encoding = tiktoken.encoding_for_model(self.model_name)
        except KeyError:
            # tiktoken only supports oai models
            # curie_logger.debug(f"Warning: model {self.model_name} not found in tiktoken. Using cl100k_base encoding.")
            encoding = tiktoken.get_encoding("cl100k_base")
        try:
            token_count = len(encoding.encode(string))
            return token_count
        except Exception as e:
            curie_logger.error(f"Error in token counting: {e}")
            return 0

    def count_message_tokens(self, message: BaseMessage) -> int:
        """Count tokens in a single message."""
        num_tokens = self.count_output_tokens(message.content)
        # Add tokens for message format (role, etc.)
        num_tokens += 4  # Format tokens
        return num_tokens

    def count_messages_tokens(self, messages: List[BaseMessage]) -> Dict[str, int]:
        """Count tokens in a list of messages."""
        input_tokens = sum(self.count_message_tokens(msg) for msg in messages)
        return {"input_tokens": input_tokens}

    def estimate_cost(self, token_counts: Dict[str, int]) -> Dict[str, float]:
        """Estimate cost based on token counts, returning costs for both input and output."""
        costs = {"input": 0.0, "output": 0.0}
        
        # Calculate input token cost
        if "input_tokens" in token_counts:
            costs["input"] = (token_counts["input_tokens"] / 1000) * self.PRICE_PER_1K_TOKENS.get(
                self.model_name, {"input": 0.01}
            )["input"]
        
        # Calculate output token cost
        if "output_tokens" in token_counts:
            costs["output"] = (token_counts["output_tokens"] / 1000) * self.PRICE_PER_1K_TOKENS.get(
                self.model_name, {"output": 0.02}
            )["output"]
        
        return costs

    def update_usage(self, token_counts: Dict[str, int]):
        """Update accumulated usage statistics."""
        # Update accumulated tokens
        self._accumulated_tokens["input"] += token_counts.get("input_tokens", 0)
        self._accumulated_tokens["output"] += token_counts.get("output_tokens", 0)
        
        # Calculate and update accumulated costs
        costs = self.estimate_cost(token_counts)
        self._accumulated_cost["input"] += costs["input"]
        self._accumulated_cost["output"] += costs["output"]

def create_model():
    return ChatLiteLLM(model=os.environ.get("MODEL")) 

def create_completion(messages: List[BaseMessage], tools: List = None) -> Any:
    """Create a completion using LiteLLM"""
    try:
        chat = ChatLiteLLM(model=os.environ.get("MODEL"))
        if tools:
            if "anthropic" in os.environ.get("MODEL"):
                chat = chat.bind_tools(tools)
            else:
                chat = chat.bind_tools(tools, parallel_tool_calls=False) # disabling for gpt models
        messages = format_messages(messages)
        return chat.invoke(messages)
    except Exception as e:
        curie_logger.error(f"Error in LLM API create_completion: {e}")
        raise e

def format_messages(messages: List[BaseMessage]) -> List[BaseMessage]:
    for msg in messages:
        if msg.content == "":
            msg.content = "Empty message."
    return messages

def update_tool_costs(costs: float):
    """Update accumulated tool usage statistics."""
    TokenCounter._accumulated_cost["tool_cost"] += costs

def get_accumulated_cost() -> float:
    """Get accumulated tool usage cost."""
    return TokenCounter._accumulated_cost 

def query_model_safe(
    messages: List[BaseMessage],
    tools: List = None,
    max_retries: int = 3,
    delay: int = 2
) -> Any:
    """Execute model query with token counting and cost estimation."""
    token_counter = TokenCounter()
    context_length = utils.get_model_context_length()
    max_tokens = context_length - 10000  # Reserve tokens for response
    
    attempt = 0
    while attempt < max_retries:
        try:
            # Count input tokens before processing
            token_counts = token_counter.count_messages_tokens(messages)
            
            # Case 1: Prune messages if total tokens exceed context length
            if token_counts["input_tokens"] > max_tokens:
                curie_logger.info(f"Total tokens ({token_counts['input_tokens']}) exceed limit ({max_tokens}). Pruning messages.")
                start = len(messages) // 3
                end = -len(messages) // 3
                messages_to_prune = messages[start:end]
                j = len(messages_to_prune) - 1
                if messages[end].type == "tool": # edge case where the first message following the last message in messages_to_prune is a tool message. So that means we DO NOT want to prune the last message in messages_to_prune. 
                    messages_to_prune = messages_to_prune[:-1]
                    end -= 1
                    j -= 1
                while j >= 0:
                    if messages_to_prune[j].type == "tool":
                        if j > 0:
                            messages_to_prune.pop()
                            messages_to_prune.pop()
                        j -= 2
                    else:
                        messages_to_prune.pop()
                        j -= 1
                messages = messages[:start] + messages_to_prune + messages[end:]
                token_counts = token_counter.count_messages_tokens(messages)
                curie_logger.info(f"After pruning - Tokens: {token_counts['input_tokens']}")

            # Case 2: Handle large last message
            last_message_tokens = token_counter.count_message_tokens(messages[-1])
            if last_message_tokens > max_tokens // 2:
                curie_logger.info(f"Last message too large ({last_message_tokens} tokens). Splitting and summarizing.")
                chunks = text_splitter_by_tokens(messages[-1].content, max_tokens // 2, token_counter)
                
                if len(chunks) > 1:
                    summarized_text = ""
                    for i, chunk in enumerate(chunks):
                        curie_logger.info(f"Processing chunk {i + 1} of {len(chunks)}")
                        summary_messages = [
                            HumanMessage(
                                content="Summarize the following text. Be concise, but maintain structure. Don't output anything other than the summarized text.\n" + chunk
                            )
                        ]
                        response = create_completion(summary_messages)
                        summarized_text += response.content + "\n"
                        curie_logger.info(f"Chunk {i + 1} summary: {response.content}")
                    
                    messages[-1].content = summarized_text.strip()

                token_counts = token_counter.count_messages_tokens(messages)
                curie_logger.info(f"After summarizing - Tokens: {token_counts['input_tokens']}")

            # Execute final completion
            response = create_completion(messages, tools=tools)
            curie_logger.debug(f"Response: {response}")

            # use tiktoken to count output tokens
            token_counts["output_tokens"] = token_counter.count_output_tokens(response.content)
            token_counter.update_usage(token_counts)            
            # Get current costs
            costs = token_counter.estimate_cost(token_counts)
            accumulated_stats = TokenCounter.get_accumulated_stats()
            # FIXME: this does not count external tool API cost 
            curie_logger.info(f"💰 Round Cost: ${sum(costs.values()):.4f} / Total Cost: ${accumulated_stats['total_cost']:.4f}")

            return response

        except BadRequestError as e:
            curie_logger.error(f"Bad request error: {e}")
            attempt += 1
            if attempt < max_retries:
                curie_logger.error(f"Retrying in {delay} seconds...")
                time.sleep(delay)
            else:
                raise e
        except Exception as e:
            curie_logger.error(f"Unexpected error: {e}")
            attempt += 1
            if attempt < max_retries:
                curie_logger.error(f"Retrying in {delay} seconds...")
                time.sleep(delay)
            else:
                raise e
    
    raise RuntimeError(f"Failed after {max_retries} retries.")

# Keep these helper functions unchanged
def text_splitter_by_tokens(text: str, chunk_size: int, token_counter: TokenCounter) -> List[str]:
    """Split text based on token count instead of characters."""
    if isinstance(text, list):
        return []

    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap=100,
        length_function=lambda x: token_counter.count_output_tokens(x),
        is_separator_regex=False,
    )
    return text_splitter.split_text(text)


# Overall guide: https://python.langchain.com/docs/how_to/structured_output/#typeddict-or-json-schema
from pydantic import BaseModel, Field
from typing import List, Dict, Any

import sys, os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from formatter import NewExperimentalPlanResponseFormatter

# store_agent = store_agent.with_structured_output(ExperimentalPlanResponseFormatter)

print("Testing valid input 1...")
valid_input = {'hypothesis': 'The c5.large instance type can handle 500 concurrent requests to the add_to_cart function with a 99th percentile latency below 100ms, while minimizing cost.', 'constant_vars': ['e-commerce web application code', 'number of concurrent requests (500)'], 'independent_vars': ['EC2 instance type'], 'dependent_vars': ['99th percentile latency', 'cost'], 'controlled_experiment_setup_description': 'Create a c5.large EC2 instance, deploy the e-commerce web application on the instance, simulate 500 concurrent requests to the add_to_cart function, measure the 99th percentile latency and cost.', 'control_group': [{'instance_type': 'c5.large'}], 'experimental_group': [{'instance_type': 'c5.medium'}], 'priority': 1}

formatter = NewExperimentalPlanResponseFormatter(**valid_input)
print(formatter.dict()) # https://docs.pydantic.dev/1.10/usage/exporting_models/

print("Testing valid input 2... (without experimental_group)")
valid_input = {'hypothesis': 'The c5.large instance type can handle 500 concurrent requests to the add_to_cart function with a 99th percentile latency below 100ms, while minimizing cost.', 'constant_vars': ['e-commerce web application code', 'number of concurrent requests (500)'], 'independent_vars': ['EC2 instance type'], 'dependent_vars': ['99th percentile latency', 'cost'], 'controlled_experiment_setup_description': 'Create a c5.large EC2 instance, deploy the e-commerce web application on the instance, simulate 500 concurrent requests to the add_to_cart function, measure the 99th percentile latency and cost.', 'control_group': [{'instance_type': 'c5.large'}], 'priority': 1}

formatter = NewExperimentalPlanResponseFormatter(**valid_input)
print(formatter.dict()) # https://docs.pydantic.dev/1.10/usage/exporting_models/


print("Testing invalid input 1... (this should trigger an error)")
invalid_input = {
    "hypothesis": "AWS EC2 VMs in us-east-1 are slower than us-east-2.",
    "constant_vars": ["EC2 VM type"],
    "independent_vars": ["AWS region"],
    "dependent_vars": ["execution_time"],
    "controlled_experiment_setup_description": "Create EC2 VM, SSH into VM, run predefined task.",
    "control_group": {"region": "us-east-1", "execution_time": "baseline"},
    "experimental_group": {"region": "us-east-2", "execution_time": "to compare"},
    "priority": 1,
}

formatter = NewExperimentalPlanResponseFormatter(**invalid_input)
print(formatter.dict()) # https://docs.pydantic.dev/1.10/usage/exporting_models/


from langchain_core.tools import tool
from typing import Annotated, List
from langgraph.store.memory import InMemoryStore
from langchain_core.messages import HumanMessage
from langgraph.prebuilt import InjectedStore
from langgraph.prebuilt import InjectedState
from langgraph.graph import END
import shutil
import os
from typing import Optional, Type, Dict
import heapq

from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field

from collections import deque, defaultdict
import re
import json
import subprocess
import logging

import model
from langchain_core.messages import HumanMessage, SystemMessage 
import formatter
import settings
import utils
from modified_deps.langchain_bash.tool import ShellTool
from logger import init_logger

class SchedNode():
    def __init__(self, store: InMemoryStore, metadata_store: InMemoryStore, State, config: dict):
        self.store = store
        self.metadata_store = metadata_store
        self.State = State
        self.sched_namespace = self.get_sched_namespace()
        self.plan_namespace = self.get_plan_namespace()
        self.config = config
        log_filename = f'../{config["log_filename"]}'
        self.curie_logger = init_logger(log_filename)
 
        self.workspace_dirname = ( self.config["workspace_name"] or 
                          self.config["job_name"] or 
                          "project" )   
        self.setup_sched()

    def setup_sched(self):
        memory_id = str("worker_assignment_dict") # Format of this dict: {"worker_name": [(exp_plan_id1, "experimental_group_partition_1"), (exp_plan_id2, "experimental_group_partition_1"), ...]}
        assignment_dict = {}
        for name in settings.list_worker_names():
            assignment_dict[name] = []
        self.metadata_store.put(self.sched_namespace, memory_id, assignment_dict)

        memory_id = str("control_worker_assignment_dict") # Format of this dict: {"worker_name": [(exp_plan_id1, "control_group"), (exp_plan_id2, "control_group"), ...]}
        assignment_dict = {}
        for name in settings.list_control_worker_names():
            assignment_dict[name] = []
        self.metadata_store.put(self.sched_namespace, memory_id, assignment_dict)

        memory_ids = {
            "llm_verifier_assignment_dict": "llm_verifier",
            "exec_verifier_assignment_dict": "exec_verifier",
            "patch_verifier_assignment_dict": "patch_verifier",
            "analyzer_assignment_dict": "analyzer",
            "concluder_assignment_dict": "concluder",
        }

        def create_assignment_dict(memory_ids, metadata_store, sched_namespace):
            for memory_id, verifier_name in memory_ids.items():
                assignment_dict = {verifier_name: []}
                metadata_store.put(sched_namespace, memory_id, assignment_dict)

        create_assignment_dict(memory_ids, self.metadata_store, self.sched_namespace)

        # Lists to track various workflow tool interactions
        write_lists = [
            # Records calls to exp_plan_write by the supervisor.
            "supervisor_wrote_list",  # [exp_plan_id1, exp_plan_id2, ...]
            # Records calls to workflow_verified_record by LLM verifier.
            "llm_verifier_wrote_list",  # [{"plan_id": ..., "partition_name": ..., "is_correct": ..., "verifier_log_message": ...}, ...]
            # Records calls to workflow_verified_record by patch verifier.
            "patch_verifier_wrote_list",  # Same format as above
            # Records calls to workflow_verified_record by analyzer.
            "analyzer_wrote_list",  # Same format as above
            # Records calls to workflow_verified_record by concluder.
            "concluder_wrote_list",  # Same format as above
            # Records calls to redo_exp_partition by supervisor.
            "supervisor_redo_partition_list",  # [{"plan_id": ..., "group": ..., "partition_name": ..., "error_feedback": ...}, ...]
            "standby_exp_plan_list",  
            "user_router_wrote_list",  
        ]


        queues = [
            "worker_queue", # Format of this queue: [(priority, exp_plan_id1, "experimental_group_partition_2"), (priority, exp_plan_id2, "experimental_group_partition_2")] # Priority queue implemented using min heap 
            "control_worker_queue"
        ]

        def create_empty_list(memory_ids, metadata_store, sched_namespace):
            for memory_id in memory_ids:
                metadata_store.put(sched_namespace, memory_id, [])

        create_empty_list(write_lists, self.metadata_store, self.sched_namespace)
        create_empty_list(queues, self.metadata_store, self.sched_namespace)
    
    def create_SchedNode_subgraph(self, sched_tool):    
        """
            No LLM involved. Manual flow control. Reason: tools cannot modify state directly (and workaround are too troublesome https://github.com/langchain-ai/langgraph/discussions/1247) which we need to control which nodes to call next. 

            Workflow:
            - Supervisor -> SchedNode -> sched_tool -> SchedNode -> Worker X (multiple are possible. We need async too TODO)
            - Worker X -> SchedNode -> sched_tool -> SchedNode -> Supervisor
            - Note that supervisor to supervisor and worker to worker transitions are technically possible too, it is up to the sched_tool. 
        """

        def SchedNode(state: self.State):
            # Invoke sched_tool: 
            response = sched_tool.invoke({"state":state}) # response is guaranteed to be a dict
            # Based on sched_tool, response, we decide which node to call next:
            if "next_agent" in response and response["next_agent"] == END: # terminate experiment. Langgraph will call END based on next_agent as defined in our conditional_edge in main.py
                self.write_down_exp_plan()
                return {"messages": state["messages"], "next_agent": END, "prev_agent": "sched_node", "remaining_steps_display": state["remaining_steps"]}
            elif "next_agent" not in response: # next agent is a worker, and we need to determine which worker it will be.
                # TODO: not doing parallelism for now, so we just assume for instances that messages will only contain one worker's name, and we will only need to call that one worker. Parallelism will be implemented later.
                control_empty = not response["control_work"]["messages"]
                experimental_empty = not response["experimental_work"]["messages"]
                assert control_empty != experimental_empty # only one of them should be non-empty. in our current iteration, since there is only one plan (I hope), this means one of these should be empty, since we cannot have that same plan not having a control group, but having some experimental groups to be run, and vice versa. TODO: thus, we will need to change this and the following lines later to accommodate more than one plan existing
                
                if response["experimental_work"]["messages"]:
                    type_name = "experimental_work"
                elif response["control_work"]["messages"]:
                    type_name = "control_work"
                assert len(list(response[type_name]["messages"].keys())) == 1 # only one worker for a given worker type has been assigned partitions to run
                self.write_down_exp_plan()
                return {"messages": [
                            HumanMessage(content=json.dumps(list(response[type_name]["messages"].values())[0]), name="scheduler")
                        ], 
                    "prev_agent": "sched_node", 
                    "next_agent": list(response[type_name]["messages"].keys())[0],
                    "remaining_steps_display": state["remaining_steps"],
                } # next_agent: worker_1
            else:
                new_dict = response.copy()
                new_dict["messages"] = [HumanMessage(content=str(response["messages"]), name="scheduler")]
                new_dict["prev_agent"] = "sched_node"
                new_dict["remaining_steps_display"] = state["remaining_steps"]
                # print(new_dict)
                # if "reset_steps" in response and response["reset_steps"] == True:
                #     new_dict["remaining_steps"] = state["remaining_steps"] + 1
                #     new_dict["remaining_steps_display"] = new_dict["remaining_steps"]
                return new_dict
        
        return SchedNode

    def write_down_exp_plan(self):

        items = self.store.search(self.plan_namespace)
        plans = [item.dict()["value"] for item in items]

        filename = self.config['exp_plan_filename'].split("/")[-1].replace(".txt", ".json")
        dirname = self.config['log_filename'].split("/")[:-1]
        with open('/' + '/'.join(dirname) + '/' + filename, 'w') as file:
            for plan in plans:
                file.write(json.dumps(plan) + "\n")

    def update_queues(
        self, 
        plan_id, 
        redo_details: Annotated[dict, "If this is not None, this means that the supervisor has requested that this partition be redone, and the error feedback is provided."]=None,
        assert_new_control: Annotated[bool, "If true, this means that the control group was just completed, meaning all experimental groups should not be completed."]=False
    ):
        """Given a plan ID, this function will:
            - if control group is done: add experimental groups that don't yet exist in the worker queue, or modify existing groups as needed. Remove plan from standy list if exist. 
            - if control group is not done: add control group that don't yet exist in the control worker queue, or modify existing control group as needed. Add plan to standby list, or modify existing as needed. 
        """
        self.curie_logger.debug("------------ Update Queues ------------")
        plan = self.store.get(self.plan_namespace, plan_id).dict()["value"]
        self.curie_logger.info(f"Plan is: {utils.pretty_json(plan)} ")

        # First, if control group is not done:
        if plan["control_group"]['partition_1']["done"] == False: # only 1 partition for now in control group
            self.curie_logger.debug("Control group is not done..")
            # Add plan to control queue if not exist or modify existing control group in queue as needed:
            partition_name = "partition_1" # Only 1 partition for now in control group
            if redo_details:
                # partition_name = redo_details["partition_name"]
                assert redo_details["group"] == "control_group"
                assert redo_details["partition_name"] == partition_name # NOTE: there is only one control group partition now, so we are certain that the error feedback will be directed to partition_1
                task_details = {
                    "priority": int(plan["priority"]),
                    "plan_id": plan_id,
                    "group": "control_group",
                    "partition_name": partition_name,
                    "workspace_dir": self.get_workspace_dirname(plan_id),
                    "error_feedback": redo_details["error_feedback"]
                }
            else:
                task_details = {
                    "priority": int(plan["priority"]),
                    "plan_id": plan_id,
                    "group": "control_group",
                    "workspace_dir": self.get_workspace_dirname(plan_id),
                    "partition_name": partition_name,
                }

            self.insert_control_worker_queue(task_details)

            # Add plan to standby list if not exist: 
            self.insert_standby_exp_plan_list(plan_id)
            self.curie_logger.debug(f"Current control group worker queue: {self.get_control_worker_queue()}")
        else: # Second, if control group is done:
            self.curie_logger.debug("Control group is done..")
            # Remove plan from standby list if exist:
            self.remove_standby_exp_plan_list(plan_id)

            # Add new experimental groups to worker queue or modify existing groups in queue as needed: 
            pq = self.metadata_store.get(self.sched_namespace, "worker_queue").dict()["value"]
            
            all_groups = self.get_groups_from_plan(plan["experimental_group"])

            self.curie_logger.debug(f"All experimental group's partitions are: {all_groups}")

            if redo_details: # if redo partition, only the partition needs to be added to queue
                task_details = {
                    "priority": int(plan["priority"]),
                    "plan_id": plan_id,
                    "group": "experimental_group",
                    "partition_name": redo_details["partition_name"],
                    "workspace_dir": self.get_workspace_dirname(plan_id),
                    "error_feedback": redo_details["error_feedback"]
                }
                self.insert_worker_queue(task_details)
            else:
                # If not redo_partition, this means that all experimental groups should be added to queue 
                for partition_name in all_groups:
                    # We do not modify partitions that are already done: (they wouldn't be in the queue anyway)
                    if plan["experimental_group"][partition_name]["done"] == True:
                        if assert_new_control:
                            raise RuntimeError("Control group just completed done, therefore no experimental groups should be done yet.")
                        continue

                    # Insert into queue:
                    task_details = {
                        "priority": int(plan["priority"]),
                        "plan_id": plan_id,
                        "group": "experimental_group",
                        "workspace_dir": self.get_workspace_dirname(plan_id),
                        "partition_name": partition_name,
                    }
                    self.insert_worker_queue(task_details)
            self.curie_logger.debug(f"Current worker queue: {self.get_worker_queue()}")
        
    def get_groups_from_plan(self, group_dict: dict) -> int:
        # Obtains group (either experimental_group or control_group) partitions from the plan
        # pattern = r"experimental_group_partition_\d+(?!_done)"

        # Collect experimental groups matches from the list
        matches = []
        for key in group_dict: # we know for sure that key will be partition_<number> (this is guaranteed by formatter)
            matches.append(key)
        return matches

    def get_worker_group_type(self, worker_name: str) -> str:
        if worker_name in settings.list_control_worker_names():
            return "control"
        elif worker_name in settings.list_worker_names():
            return "experimental"
        else:
            raise ValueError("Worker name not found in list of workers.")
    
    def has_idle_worker(self, group_type: str) -> (bool, str):
        memory_id = self.get_assignment_dict_mem_id(group_type)

        assignment_dict = self.metadata_store.get(self.sched_namespace, memory_id).dict()["value"]
        for name in assignment_dict:
            if not assignment_dict[name] or len(assignment_dict[name]) < settings.PARTITIONS_PER_WORKER:
                return True, name
        return False, None

    def assign_worker(self, group_type: str) -> dict:
        assignment_messages = defaultdict(list) # {worker_1: [{plan_id: xyz, partition_name: xyz}], worker_2: [{plan_id: xyz, partition_name: xyz}]}
        while True: # Keep assigning until no more idle workers
            has_idle_worker, worker_name = self.has_idle_worker(group_type)
            if has_idle_worker:
                task_details = self.pop_worker_queue(group_type)
                if task_details: # if queue is not empty
                    priority, plan_id, group, partition_name = task_details["priority"], task_details["plan_id"], task_details["group"], task_details["partition_name"]
                    self._assign_worker(worker_name, task_details, group_type)
                    if "error_feedback" in task_details:
                        task_details["error_feedback"] = self.augment_redo_partition_error_feedback(task_details)
                    assignment_messages[worker_name].append(task_details) # only the worker itself needs to be passed in context that may include an error feedback
                else:
                    break # no more plans in queue
            else:
                break

        return assignment_messages

    def augment_redo_partition_error_feedback(self, task_details: dict) -> str:
        self.curie_logger.info("------------Entering augment redo partition error feedback!!!------------")
        plan_id = task_details["plan_id"]
        group = task_details["group"]
        partition_name = task_details["partition_name"]
        error_feedback = task_details["error_feedback"]
        # Get plan from plan_id:
        plan = self.store.get(self.plan_namespace, plan_id).dict()["value"]
        # Get control experiment filename:
        control_experiment_filename = plan[group][partition_name]["control_experiment_filename"]
        self.curie_logger.info("------------Exiting augment redo partition error feedback!!!------------")
        # Return augmented error feedback:
        if not control_experiment_filename:
            return error_feedback
        else:
            return error_feedback + " Consider reusing the existing experiment workflow setup that you made earlier as your starting point, but feel free to start from scratch if you believe it can't be fixed or salvaged. The workflow filename is: {}".format(control_experiment_filename)

    
    def _assign_worker(self, worker_name: str, assignment_dict: dict, group_type: str):
        memory_id = self.get_assignment_dict_mem_id(group_type)
        self._assign_to_entity(worker_name, assignment_dict, memory_id)
    
    def assign_verifier(self, verifier_name, assignment_dict: dict):
        memory_id = self.get_assignment_dict_mem_id(verifier_name)
        self._assign_to_entity(verifier_name, assignment_dict, memory_id)

    def _assign_to_entity(self, entity_name: str, assignment_dict: dict, memory_id: str):
        overall_assignment_dict = self.metadata_store.get(self.sched_namespace, memory_id).dict()["value"]
        overall_assignment_dict[entity_name].append(assignment_dict)
        self.metadata_store.put(self.sched_namespace, memory_id, overall_assignment_dict)

    def get_worker_assignment(self, worker_name: str) -> list:
        group_type = self.get_worker_group_type(worker_name)
        memory_id = self.get_assignment_dict_mem_id(group_type)
        return self._get_entity_assignment(worker_name, memory_id)
    
    def get_verifier_assignment(self, verifier_name: str) -> list:
        memory_id = self.get_assignment_dict_mem_id(verifier_name)
        return self._get_entity_assignment(verifier_name, memory_id) # format: [(plan_id1, "experimental_group", "partition_1"), (plan_id2, "experimental_group", "partition_1"), ...]

    def _get_entity_assignment(self, entity_name, memory_id):
        assignment_dict = self.metadata_store.get(self.sched_namespace, memory_id).dict()["value"]
        return assignment_dict[entity_name]
    
    def unassign_worker_all(self, worker_name: str):
        """ Unassign all groups from a worker. """
        group_type = self.get_worker_group_type(worker_name)
        memory_id = self.get_assignment_dict_mem_id(group_type)
        self._unassign_entity_all(worker_name, memory_id)
    
    def unassign_verifier_all(self, verifier_name: str):
        memory_id = self.get_assignment_dict_mem_id(verifier_name)
        self._unassign_entity_all(verifier_name, memory_id)
    
    def _unassign_entity_all(self, entity_name: str, memory_id: str):
        assignment_dict = self.metadata_store.get(self.sched_namespace, memory_id).dict()["value"]
        assignment_dict[entity_name] = []
        self.metadata_store.put(self.sched_namespace, memory_id, assignment_dict)

    def pop_worker_queue(self, group_type: str) -> (dict):
        # https://langchain-ai.github.io/langgraph/reference/store/#langgraph.store.base.BaseStore.get
        if group_type == "experimental":
            pq = self.metadata_store.get(self.sched_namespace, "worker_queue").dict()["value"]
        elif group_type == "control":
            pq = self.metadata_store.get(self.sched_namespace, "control_worker_queue").dict()["value"]
        if pq:
            _, _, _, _, task_details = heapq.heappop(pq)
            if group_type == "experimental":
                self.metadata_store.put(self.sched_namespace, "worker_queue", pq)
            elif group_type == "control":
                self.metadata_store.put(self.sched_namespace, "control_worker_queue", pq)
            return task_details
        else:
            return None
    
    def insert_worker_queue(self, task_details: dict):
        pq = self.metadata_store.get(self.sched_namespace, "worker_queue").dict()["value"]

        priority, plan_id, group, partition_name = task_details["priority"], task_details["plan_id"], task_details["group"], task_details["partition_name"]

        # Check if plan is in queue (linear time). If plan exists, remove it from queue. Note we need to do this because there could be a change in priority even if the plan is already in the queue. 
        for index in range(len(pq)):
            _, _, _, _, task_details2 = pq[index]
            priority2, plan_id2, group2, partition_name2 = task_details2["priority"], task_details2["plan_id"], task_details2["group"], task_details2["partition_name"]
            if plan_id == plan_id2 and group == group2 and partition_name == partition_name2:
                pq.pop(index)
                break # there should be only one such an instance anyway

        heapq.heappush(pq, (priority, plan_id, group, partition_name, task_details)) # heapq needs priority and the other fields except for task_details since its a dict, to make comparison
        self.metadata_store.put(self.sched_namespace, "worker_queue", pq)

    def insert_control_worker_queue(self, task_details: dict):
        pq = self.metadata_store.get(self.sched_namespace, "control_worker_queue").dict()["value"] # format: [(priority, plan_id1)]

        priority, plan_id, group, partition_name = task_details["priority"], task_details["plan_id"], task_details["group"], task_details["partition_name"]

        # Check if plan is in queue (linear time). If plan exists, remove it from queue. Note we need to do this because there could be a change in priority even if the plan is already in the queue. 
        for index in range(len(pq)):
            _, _, _, _, task_details2 = pq[index]
            priority2, plan_id2, group2, partition_name2 = task_details2["priority"], task_details2["plan_id"], task_details2["group"], task_details2["partition_name"]
            if plan_id == plan_id2 and group == group2 and partition_name == partition_name2:
                pq.pop(index)
                break # there should be only one such an instance anyway

        heapq.heappush(pq, (priority, plan_id, group, partition_name, task_details))
        self.metadata_store.put(self.sched_namespace, "control_worker_queue", pq)

    def get_worker_queue(self):
        pq = self.metadata_store.get(self.sched_namespace, "worker_queue").dict()["value"]
        return pq

    def get_control_worker_queue(self):
        pq = self.metadata_store.get(self.sched_namespace, "control_worker_queue").dict()["value"] # format: [(priority, plan_id1)]
        return pq

    def write_all_control_experiment_results_filenames(self, completion_messages: list):
        # for the format of completion_messages, see tool.LLMVerifierWriteTool for the format of this list
        for item in completion_messages:
            plan_id, group, partition_name = item["plan_id"], item["group"], item["partition_name"]
            plan = self.store.get(self.plan_namespace, plan_id).dict()["value"]
            filename = self.get_all_control_experiment_results_filename(plan_id, group, partition_name)
            with open(filename, 'w') as file:
                file.write(item["verifier_log_message"])
            plan[group][partition_name]["all_control_experiment_results_filename"] = filename
            self.store.put(self.plan_namespace, plan_id, plan)

    def insert_standby_exp_plan_list(self, plan_id: str):
        standby_exp_plan_list = self.metadata_store.get(self.sched_namespace, "standby_exp_plan_list").dict()["value"]

        # Check if plan is in queue (linear time). If plan exists, remove it from queue. Note we need to do this because there could be a change in priority even if the plan is already in the queue. 
        for index in range(len(standby_exp_plan_list)):
            plan_id2 = standby_exp_plan_list[index]
            if plan_id == plan_id2:
                standby_exp_plan_list.pop(index)
                break # there should be only one such an instance anyway

        standby_exp_plan_list.append(plan_id)
        self.metadata_store.put(self.sched_namespace, "standby_exp_plan_list", standby_exp_plan_list)

    def remove_standby_exp_plan_list(self, plan_id: str):
        standby_exp_plan_list = self.metadata_store.get(self.sched_namespace, "standby_exp_plan_list").dict()["value"]

        # Check if plan is in list (linear time). If plan exists, remove it from list. 
        for index in range(len(standby_exp_plan_list)):
            plan_id2 = standby_exp_plan_list[index]
            if plan_id == plan_id2:
                standby_exp_plan_list.pop(index)
                break
        
        self.metadata_store.put(self.sched_namespace, "standby_exp_plan_list", standby_exp_plan_list)

    def get_verifier_wrote_list_item(self, verifier_name:str, plan_id: str, group: str, partition_name: str):
        memory_id = self.get_wrote_list_mem_id(verifier_name)

        verifier_wrote_list = self.metadata_store.get(self.sched_namespace, memory_id).dict()["value"]
        for item in verifier_wrote_list:
            if item["plan_id"] == plan_id and item["group"] == group and item["partition_name"] == partition_name:
                return item
        return None

    def get_concluder_wrote_list_item(self):
        memory_id = str("concluder_wrote_list")
        verifier_wrote_list = self.metadata_store.get(self.sched_namespace, memory_id).dict()["value"]
        return verifier_wrote_list

    def remove_verifier_wrote_list_all(self, verifier_name: str):
        memory_id = self.get_wrote_list_mem_id(verifier_name)
        self.metadata_store.put(self.sched_namespace, memory_id, [])
    
    def get_control_experiment_filename(self, plan_id: str, group: str, partition_name: str) -> str:
        return "/workspace/{}_{}/control_experiment_{}_{}_{}.sh".format(os.path.basename(self.workspace_dirname) , plan_id, plan_id, group, partition_name)

    def get_control_experiment_results_filename(self, plan_id: str, group: str, partition_name: str) -> str:
        return "/workspace/{}_{}/results_{}_{}_{}.txt".format(os.path.basename(self.workspace_dirname), plan_id, plan_id, group, partition_name)

    def get_all_control_experiment_results_filename(self, plan_id: str, group: str, partition_name: str) -> str:
        # results for multiple runs (i.e., a single run by exec verifier for now) for a single partition
        return "/workspace/{}_{}/all_results_{}_{}_{}.txt".format(os.path.basename(self.workspace_dirname) , plan_id, plan_id, group, partition_name)

    def get_workspace_dirname(self, plan_id: str) -> str: 
        return "/workspace/{}_{}".format(os.path.basename(self.workspace_dirname), plan_id) 

    def is_no_plan_exists(self):
        items = self.store.search(self.plan_namespace)
        self.curie_logger.debug(f"Plans that exist: {items}, with len: {len(items)}")
        plans_list = [item.dict()["value"] for item in items]
        if len(plans_list) == 0:
            return True
        else:
            return False

    def check_exp_termination_condition(self):
        """
        If all control and experimental groups are done for all plans, return True. Otherwise, return False.
        """
        items = self.store.search(self.plan_namespace)

        plans_list = [item.dict()["value"] for item in items]

        for plan in plans_list:
            for partition_name in plan["control_group"]:
                if plan["control_group"][partition_name]["done"] == False:
                    return False
            for partition_name in plan["experimental_group"]:
                if plan["experimental_group"][partition_name]["done"] == False:
                    return False
        return True

    def get_sched_namespace(self):
        user_id = "admin"
        application_context = "exp-sched" 
        sched_namespace = (user_id, application_context) # just a random namespace name for now
        return sched_namespace
    
    def get_plan_namespace(self):
        user_id = "admin"
        application_context = "exp-plans" 
        plan_namespace = (user_id, application_context) # just a random namespace name for now
        return plan_namespace

    def get_wrote_list_mem_id(self, verifier_name: str) -> str:
        if verifier_name == "llm_verifier":
            memory_id = str("llm_verifier_wrote_list") # check tool.LLMVerifierWriteTool for the format of this list
        elif verifier_name == "patch_verifier":
            memory_id = str("patch_verifier_wrote_list")
        elif verifier_name == str("analyzer"):
            memory_id = str("analyzer_wrote_list")
        elif verifier_name == str("concluder"):
            memory_id = str("concluder_wrote_list")
        return memory_id
    
    def get_assignment_dict_mem_id(self, entity_name: str) -> str:
        if entity_name == "llm_verifier":
            memory_id = "llm_verifier_assignment_dict"
        elif entity_name == "exec_verifier":
            memory_id = "exec_verifier_assignment_dict"
        elif entity_name == "patch_verifier":
            memory_id = "patch_verifier_assignment_dict"
        elif entity_name == "analyzer":
            memory_id = "analyzer_assignment_dict"
        elif entity_name == "concluder":
            memory_id = "concluder_assignment_dict"
        elif entity_name == "experimental":
            memory_id = str("worker_assignment_dict")
        elif "control" in entity_name:
            memory_id = str("control_worker_assignment_dict")
        else:
            raise ValueError(f"Entity name not found for {entity_name}")
        return memory_id
    
    def copy_dataset_to_workspace(self):
        if self.config["dataset_dir"] != "": 
            # Copy dataset to workspace
            dataset_dir = os.path.join('/all', self.config["dataset_dir"].lstrip('/').rstrip('/'))

            if not os.path.exists(dataset_dir):
                raise FileNotFoundError(f"Dataset directory does not exist: {self.config['dataset_dir']}. Please check the path.")            

            workspace_dir = "/workspace/" 
            dataset_dir_name = (os.path.basename(self.config["workspace_name"]) or 
                                os.path.basename(self.config["job_name"]) or
                                "project" ) + "_dataset"
            new_dataset_dir = os.path.join(workspace_dir, dataset_dir_name)
            dataset_name = dataset_dir.split("/")[-1]

            if os.path.isfile(dataset_dir):
                dataset_dir_size = os.path.getsize(dataset_dir)
            else:
                dataset_dir_size = sum(os.path.getsize(os.path.join(dataset_dir, f)) for f in os.listdir(dataset_dir) if os.path.isfile(os.path.join(dataset_dir, f)))

            if os.path.exists(new_dataset_dir):
                if os.path.isfile(new_dataset_dir):
                    new_dataset_dir_size = os.path.getsize(new_dataset_dir)
                else:
                    new_dataset_dir_size = sum(os.path.getsize(os.path.join(new_dataset_dir, f)) for f in os.listdir(new_dataset_dir) if os.path.isfile(os.path.join(new_dataset_dir, f)))
            else:
                new_dataset_dir_size = 0
            
            if os.path.isfile(dataset_dir):
                if os.path.exists(new_dataset_dir):
                    shutil.rmtree(new_dataset_dir)

                # mkdir new_dataset_dir
                os.makedirs(new_dataset_dir, exist_ok=True)
                shutil.copy(dataset_dir, new_dataset_dir)
                self.curie_logger.info(f"Copying file {dataset_dir} --> {new_dataset_dir} successfully!") 

            elif os.path.isdir(dataset_dir):
                if not os.path.exists(new_dataset_dir) or ( os.path.exists(new_dataset_dir) and dataset_dir_size != new_dataset_dir_size):
                    if os.path.exists(new_dataset_dir): 
                        shutil.rmtree(new_dataset_dir) 

                    try:
                        subprocess.run(["cp", "-r", f"{dataset_dir}",  '/workspace'], check=True)
                        os.rename(os.path.join('/workspace', dataset_name), new_dataset_dir)
                        self.curie_logger.info(f"Copying {dataset_dir} --> {new_dataset_dir} successfully!") 
                    except Exception as e:
                        self.curie_logger.info(f"Error copying files: {e}")
                        raise
                else:
                    self.curie_logger.info(f"Dataset directory already exists: {new_dataset_dir}. Skipping copy.")

            return new_dataset_dir
        else:
            self.curie_logger.info(f"No dataset directory specified. Skipping dataset copy.")
            
    def init_new_plan(self, plan_ids: list):
        import concurrent.futures

        new_dataset_dir = self.copy_dataset_to_workspace()
        self.get_packages_to_install()

        def process_plan(plan_id):
            # FIXME: initialization time too long, trigger this when deciding to execute this plan
            self.curie_logger.info(f"Preparing environment for plan ID: {plan_id}")
            new_plan = self.create_workspace_dir(plan_id)
            if new_plan:
                # Add "workspace_dir" attribute to plan
                self.add_workspace_to_plan(plan_id)
                # Edit plan question:
                self.edit_plan_question(plan_id)
                self.add_dataset_to_plan(plan_id, new_dataset_dir)
            return plan_id
        
        # Execute the plans in parallel using a ThreadPoolExecutor
        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Submit all tasks and get future objects
            futures = [executor.submit(process_plan, plan_id) for plan_id in plan_ids]
            
            # Wait for all futures to complete (optional)
            concurrent.futures.wait(futures)

    def init_coding_env(self, work_dir: str, env_name: str='venv'):
        
        def install_packages(env_path, packages):
            """
            Install specified packages into a micromamba environment
            
            Args:
                env_path (str): Path to the micromamba environment
                packages (list): List of packages to install
            """
            if not os.path.exists(env_path) or len(packages) == 0:
                return 
            # Activate the environment
            successful_packages = []
            failed_packages = []
            # start_time = time.time()
            for i, package in enumerate(packages):
                if package == "time":
                    continue
                # Construct the installation command for the current package
                activate_cmd = [
                    "micromamba", "install", "-y", "--quiet",
                    "-p", env_path, package, '&'
                ]
                try:
                    # Run the installation command for the current package
                    result = subprocess.run(
                        activate_cmd, 
                        check=True, 
                        stdout=subprocess.PIPE, 
                        stderr=subprocess.PIPE, 
                        text=True
                    )
                    successful_packages.append(package)
                    self.curie_logger.info(f"[{i}/{len(packages)}] Successfully installed {package}.")

                except subprocess.CalledProcessError as e:
                    try:
                        activate_cmd = [
                            "micromamba", "run", "-p", env_path,
                            "pip", "install", package
                        ]
                        result = subprocess.run(
                            activate_cmd, 
                            check=True, 
                            stdout=subprocess.PIPE, 
                            stderr=subprocess.PIPE, 
                            text=True
                        )
                        successful_packages.append(package)
                    except subprocess.CalledProcessError as e: 
                        failed_packages.append(package) 
                        self.curie_logger.info(f"Fail to install the packages {package}. Error: {e.stderr}")
    
            # self.curie_logger.info(f"Sucessfully install packages: {', '.join(successful_packages)} in {time.time() - start_time:.2f} seconds.")
            self.curie_logger.info(f"Sucessfully install packages: {', '.join(successful_packages)}.")

        # FIXME: some use cases may need old versions of Python 
        env_path = os.path.join(work_dir, env_name)
        if not os.path.exists(env_path):
            command = ["micromamba", "create", "-p", env_path, "python=3.12", "-y", "--quiet"]
            subprocess.run(command, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            try:
                # Install the packages
                install_packages(env_path, self.packages_to_install)
            except json.JSONDecodeError as e:
                self.curie_logger.info(f"No python package needs to be installed") 
        else:
            self.curie_logger.info(f"Environment is pre-built at {env_path}. Skipping creation.")

        return env_path

    def get_packages_to_install(self):
        with open('/'+self.config['exp_plan_filename'], "r") as file:  
            question = file.read() 
        with open("/curie/prompts/exp-env-manager.txt", "r") as file:
            system_prompt = file.read() 
            
        messages = [SystemMessage(content=system_prompt),
                HumanMessage(content=question)]
        response = model.query_model_safe(messages) 
        cleaned_response = response.content.replace('```json', '').replace('```', '').strip()
        try:
            self.packages_to_install  = json.loads(cleaned_response)["packages"] 
        except json.JSONDecodeError as e:
            self.curie_logger.info(f"No python package needs to be installed") 
            self.packages_to_install = []
    
    def create_workspace_dir(self, plan_id: str):
        # If we are running a question from Curie benchmark (specified in config["workspace_name"]), copy its associated starter files from ../starter_file and move it to ../workspace. 
        # Otherwise, if running a question not from Curie benchmark, we assume that starter_file does not exist, and we do not copy. We still create the new_starter_file_dir folder but leave it empty. 
        
        workspace_base_name = (os.path.basename(self.config["workspace_name"]) or
                               os.path.basename(self.config["job_name"]) or
                               "project" )
        new_starter_file_dir = f"../workspace/{workspace_base_name}_{plan_id}"
        if self.config["workspace_name"] != "":
            old_starter_file_dir = os.path.join('/all', self.config["workspace_name"].lstrip('/')) 
            new_starter_file_dir = os.path.abspath(new_starter_file_dir)  
        else:
            new_starter_file_dir = os.path.abspath(new_starter_file_dir)  
            old_starter_file_dir = None
        
        if not os.path.exists(new_starter_file_dir):
            
            try:
                if old_starter_file_dir and os.path.exists(old_starter_file_dir): 
                    # This will copy only the contents of old_starter_file_dir into new_starter_file_dir, not the directory itself.
                    subprocess.run(["cp", "-r", old_starter_file_dir,  new_starter_file_dir], check=True)

                    self.curie_logger.info(f"Created 📁 {new_starter_file_dir}. \
                                            Starter files from {old_starter_file_dir.replace('/all/', '')} copied successfully!")
                else:
                    self.curie_logger.info(f"Created 📁 {new_starter_file_dir}. No starter files to copy.")
                # FIXME: install environment for each plan_id -- too slow.
                env_path = self.init_coding_env(new_starter_file_dir)
                self.curie_logger.info(f"Micromamba environment created at {env_path}")

            except Exception as e:
                self.curie_logger.info(f"Error copying files: {e}")
                raise

            return True
        else:
            return False
        
    @staticmethod
    def write_at_beginning(filename, text_to_add): 
        try:
            with open(filename, "r") as file:
                existing_content = file.read()
        except FileNotFoundError:
            existing_content = ""   

        with open(filename, "w") as file:
            file.write(text_to_add)
            file.write(existing_content)

    def add_dataset_to_plan(self, plan_id: str, new_dataset_dir: str):
        # for plan_id in plan_id_list:
        plan = self.store.get(self.plan_namespace, plan_id).dict()["value"]
        plan["dataset_dir"] = new_dataset_dir
        self.store.put(self.plan_namespace, plan_id, plan) 
        description_file = os.path.join(plan["workspace_dir"], "description.md") 
        self.write_at_beginning(description_file,f"\nDataset directory: {plan['dataset_dir']}. \
                                Briefly explore the directory `ls * | head -n 50` to understand the structure.\n")
        
    def add_workspace_to_plan(self, plan_id: str):
        plan = self.store.get(self.plan_namespace, plan_id).dict()["value"]
        plan["workspace_dir"] = self.get_workspace_dirname(plan_id)
        self.store.put(self.plan_namespace, plan_id, plan)

    def edit_plan_question(self, plan_id: str):
        """
            Edit plan question to point to the correct writable workspace directory, that the technician agents are able to tweeak/modify. 
        """
        plan = self.store.get(self.plan_namespace, plan_id).dict()["value"]

        plan["question"] = plan["question"].replace(f"/starter_file/{self.config['workspace_name']}", self.get_workspace_dirname(plan_id))

        self.store.put(self.plan_namespace, plan_id, plan)

    def get_question(self):
        memory_id = str("question")
        question = self.metadata_store.get(self.sched_namespace, memory_id)
        question = question.dict()["value"]
        return question

    def get_concluder_terminate_message(self):
        return "Your task: Some results are incomplete, but you must conclude the experiment now (make sure to set 'is_conclude' to True; 'concluder_log_message' must also make it clear that we must conclude now). Analyze the results per the instructions and provide a detailed breakdown of what the user could do on their own (e.g. propose follow-up experiments)."

class SchedInput(BaseModel):
    state: Annotated[dict, InjectedState] # For scheduler, we are guaranteed that the prev_agent will either be supervisor or worker. 

class SchedTool(BaseTool):
    name: str = "scheduler"
    description: str = "Programmatic scheduling."
    args_schema: Type[BaseModel] = SchedInput
    # None of the following work:
    # https://langchain-ai.github.io/langgraph/how-tos/pass-run-time-values-to-tools/#define-the-tools_1
    # https://github.com/langchain-ai/langchain/discussions/24906
    # and so on..
    config: Optional[dict] = None
    curie_logger: Optional[logging.Logger] = None

    def __init__(self, store: InMemoryStore, metadata_store: InMemoryStore, config: dict):
        super().__init__()
        self.config = config
        log_filename = f'../{config["log_filename"]}'
        self.curie_logger = init_logger(log_filename)
    
    class Config:
        arbitrary_types_allowed = True  # Allow non-Pydantic types like InMemoryStore

    def _run(
        self,
        state: Annotated[dict, InjectedState],
        run_manager: Optional[CallbackManagerForToolRun] = None
    ) -> str: 
        prev_agent = state["prev_agent"]
        if state["remaining_steps"] <= settings.CONCLUDER_BUFFER_STEPS:
            if prev_agent != "concluder":
                self.curie_logger.info("------------ Not enough remaining steps to continue with experimentation. Entering concluder now!------------")
                return {"messages": [self.get_concluder_terminate_message()], "prev_agent": "analyzer", "next_agent": "concluder"} 
            else:
                self.curie_logger.info("------------ Not enough remaining steps to continue with experimentation. Handling concluder output before EXITING!!!------------")
        next_agent = self._handle_agent(prev_agent, state)
        next_agent_name = 'N/A'
        for k, v in next_agent.items():
            if k == 'next_agent':
                next_agent_name = v
                break
            if 'next_agent' in v:
                next_agent_name = v['next_agent']
                break
        
        self.curie_logger.info(
            f"<<<<<<<< Scheduling {prev_agent} ⏩ {next_agent_name} >>>>>>>>"
        )
        if next_agent_name == 'N/A':
            self.curie_logger.info(
                f"------- No next agent found for {next_agent} ---------"
            )
        
        return next_agent
    
    # TODO: currently assumes single nodes for each node type. Need to account for multiple nodes of same type.
    def _handle_agent(self, prev_agent: str, state: Annotated[dict, InjectedState]) -> str:
        """
        Route to appropriate handler based on previous agent.
        
        Args:
            prev_agent (str): Name of the previous agent
            
        Returns:
            str: Response from the appropriate handler
        """
        handlers = self.config["transition_funcs"]
        # print(handlers)
        
        # Handle special case for worker which has a prefix
        if "control_worker" in prev_agent:
            return handlers["control_worker"]()
        if "worker" in prev_agent:
            return handlers["worker"]()
        if "concluder" == prev_agent:
            return handlers["concluder"](state)
        if "supervisor" == prev_agent:
            return handlers["supervisor"](state)
        if "user_input_router" == prev_agent:
            return handlers["user_input_router"](state)
        if "user_input" == prev_agent:
            return handlers["user_input"](state)
        
        # Get the appropriate handler or raise an error if not found
        handler = handlers.get(prev_agent)
        if not handler:
            raise ValueError(f"No handler found for agent: {prev_agent}")
            
        return handler()
    
    def get_concluder_terminate_message(self):
        return "Your task: Some results are incomplete, but you must conclude the experiment now (make sure to set 'is_conclude' to True; 'concluder_log_message' must also make it clear that we must conclude now). Analyze the results per the instructions and provide a detailed breakdown of what the user could do on their own (e.g. propose follow-up experiments)."


from pydantic import BaseModel, Field
from typing_extensions import TypedDict 
from langgraph.graph import StateGraph, START, END
from langgraph.graph.message import add_messages
from langgraph.prebuilt import ToolNode, tools_condition
from langchain_core.messages import HumanMessage, SystemMessage, ToolMessage
from abc import ABC, abstractmethod

import os
import json

import model
import utils
import tool
import settings
from logger import init_logger
from scheduler import SchedNode

class NodeConfig(BaseModel):
    name: str = Field(
        description="The name of the node, used for logging and debugging."
    )

    node_icon: str = Field( 
        default="👑",
        description="The icon for the node."
    )

    log_filename: str = Field(
        description="The filename for the log file."
    )

    config_filename: str = Field(
        description="The filename for the config file."
    )

    transition_objs: dict = Field(
        default_factory=dict,
        description="A dictionary of transition objects for the node."
    )
    
    system_prompt_key: str = Field(
        default="supervisor_system_prompt_filename",
        description="The key for the system prompt in the config file."
    )

    default_system_prompt_filename: str = Field(
        default="prompts/exp-supervisor.txt",
        description="The default filename for the system prompt."
    )

class BaseNode(ABC):
    def __init__(self, sched_node: SchedNode, config: NodeConfig, State, store, metadata_store, memory, tools: list):
        self.node_config = config
        self.sched_node = sched_node # we will be using helper functions from scheduler within transition_handle_func
        self.curie_logger = init_logger(self.node_config.log_filename)
        self.State = State
        self.store = store
        self.metadata_store = metadata_store
        self.memory = memory
        self.tools = tools
        self.sched_namespace = self.sched_node.sched_namespace
        self.plan_namespace = self.sched_node.plan_namespace
    
    def get_name(self):
        return self.node_config.name

    def create_subgraph(self):
        """ Creates a Node subgraph."""

        subgraph_builder = StateGraph(self.State)
        
        with open(self.node_config.config_filename, 'r') as file:
            config = json.load(file)
    
        system_prompt_file = config.get(self.node_config.system_prompt_key, self.node_config.default_system_prompt_filename)

        subgraph_node = self._create_model_response(system_prompt_file) 

        subgraph_builder.add_node(self.node_config.name, subgraph_node)
        subgraph_builder.add_edge(START, self.node_config.name)
        tool_node = ToolNode(tools=self.tools)
        subgraph_builder.add_node("tools", tool_node)

        subgraph_builder.add_conditional_edges(self.node_config.name, tools_condition)
        # supervisor_builder.add_conditional_edges( "tools", router, ["supervisor", END])
        subgraph_builder.add_edge("tools", self.node_config.name)

        subgraph = subgraph_builder.compile(checkpointer=self.memory) 

        def call_subgraph(state: self.State) -> self.State: 
            response = subgraph.invoke({
                    "messages": state["messages"][-1]
                },
                {
                    # "recursion_limit": 20,
                    "configurable": {
                        "thread_id": f"{self.node_config.name}_graph_id"
                    }
                }
            )
            
            return {
                "messages": [
                    HumanMessage(content=response["messages"][-1].content, name=f"{self.node_config.name}_graph")
                ],
                "prev_agent": response["prev_agent"],
                "remaining_steps_display": state["remaining_steps"],
            }
        return call_subgraph

    def _create_model_response(self, system_prompt_file):    
        # FIXME: better way to get model names; from config?
        # FIXME: can move model name to model.py 
        def Node(state: self.State):
            if state["remaining_steps"] <= settings.CONCLUDER_BUFFER_STEPS:
                return {
                    "messages": [], 
                    "prev_agent": self.node_config.name,
                }
                
            # Read from prompt file:
            with open(system_prompt_file, "r") as file:
                system_prompt = file.read()

            system_message = SystemMessage(
                content=system_prompt,
            )

            # can filter out early tool messages from state["messages"]
            # need to retain the plan messages from state["messages"]
            # remove the duplicated human messages from state["messages"]
            
            # TODO: check for high similarity between messages, repeatly summarize the messages
            # If there are too many messages, prune older ToolMessages to avoid context overflow
            messages = state["messages"]
            # unique_msg_contents = {} # content -> index
            if len(messages) > 50:
                # Keep track of tool messages to potentially remove
                tool_messages = []
                to_remove = set()
                
                for i, msg in enumerate(messages):
                    # prune the first half of the messages
                    if i < min(20, len(messages) // 4):
                        continue
                    if len(messages) - i < min(30, len(messages) // 3):
                        break

                    # content = msg.content
                    # # remove duplicate messages
                    # if content not in unique_msg_contents:
                    #     unique_msg_contents[content] = i
                    # else:
                    #     to_remove.add(unique_msg_contents[content])
                    #     unique_msg_contents[content] = i

                    if isinstance(msg, ToolMessage):
                        tool_messages.append(i)
                        tool_messages.append(i-1) # corresponding ai message

                if tool_messages:
                    to_remove.update(tool_messages)
                    
                filtered_messages = [msg for i, msg in enumerate(messages) if i not in to_remove] 
            else:
                filtered_messages = messages
            

            # self.curie_logger.info(f"❕❕❕ before filtering (len: {len(state['messages'])} messages): {state['messages']}")
            # self.curie_logger.info(f"❕❕❕ after filtering (len: {len(filtered_messages)} messages ): {filtered_messages}")
            self.curie_logger.debug(f"🏦 {self.node_config.node_icon} number of saved messages: {len(state['messages'])} --> {len(filtered_messages)}")

            state["messages"] = filtered_messages
            messages = state["messages"]

            # Ensure the system prompt is included at the start of the conversation
            if not any(isinstance(msg, SystemMessage) for msg in messages):
                messages.insert(0, system_message)

            self.curie_logger.debug(f"Messages TO {self.node_config.name.upper()} {self.node_config.node_icon}: {messages}")
            
            response = model.query_model_safe(messages, tools=self.tools)

            self.curie_logger.info(f"<><><><><> {self.node_config.node_icon} {self.node_config.name.upper()} {self.node_config.node_icon} <><><><><>")

            if response.tool_calls:
                self.curie_logger.info(f"Tool calls: {response.tool_calls[0]['name']}")
                if 'prompt' in response.tool_calls[0]['args']:
                    self.curie_logger.info(f"Message received: {response.tool_calls[0]['args']['prompt']}")
                elif 'verifier_log_message' in response.tool_calls[0]['args']:
                    self.curie_logger.info(f"Message: {response.tool_calls[0]['args']['verifier_log_message']}")
                else:
                    self.curie_logger.info(f"Message: {response.tool_calls[0]['args']}")
            
            concise_msg = response.content.split('\n\n')[0]
            if self.node_config.name.upper() == "CONCLUDER":
                self.curie_logger.info(f"✌️ Concluder response: {response.content}")
            elif concise_msg:
                self.curie_logger.info(f'Concise response: {concise_msg}')

            self.curie_logger.debug(f"Full response from {self.node_config.name.upper()} {self.node_config.node_icon}: {response}")

            return {"messages": [response], "prev_agent": self.node_config.name}
            # need to change 'add_messages' if you want to permanently update the message state.
        
        return Node

    @abstractmethod
    def transition_handle_func(self):
        """Handles transition logic and determines next action to take."""
        pass

    @abstractmethod
    def create_transition_objs(self):
        """Creates transition objects for the node."""
        pass